#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import os

import requests
from hopsworks_common.client import auth, exceptions
from hopsworks_common.client.istio import base as istio


class Client(istio.Client):
    REQUESTS_VERIFY = "REQUESTS_VERIFY"
    PROJECT_ID = "HOPSWORKS_PROJECT_ID"
    PROJECT_NAME = "HOPSWORKS_PROJECT_NAME"
    HADOOP_USER_NAME = "HADOOP_USER_NAME"
    HDFS_USER = "HDFS_USER"

    DOMAIN_CA_TRUSTSTORE_PEM = "DOMAIN_CA_TRUSTSTORE_PEM"
    MATERIAL_DIRECTORY = "MATERIAL_DIRECTORY"
    T_CERTIFICATE = "t_certificate"
    K_CERTIFICATE = "k_certificate"
    TRUSTSTORE_SUFFIX = "__tstore.jks"
    KEYSTORE_SUFFIX = "__kstore.jks"
    PEM_CA_CHAIN = "ca_chain.pem"
    CERT_KEY_SUFFIX = "__cert.key"
    MATERIAL_PWD = "material_passwd"
    SECRETS_DIR = "SECRETS_DIR"

    def __init__(self, host, port):
        """Initializes a client being run from a job/notebook directly on Hopsworks."""
        self._host = host
        self._port = port
        self._base_url = "http://" + self._host + ":" + str(self._port)

        trust_store_path = self._get_ca_chain_path()
        hostname_verification = (
            os.environ[self.REQUESTS_VERIFY]
            if self.REQUESTS_VERIFY in os.environ
            else "true"
        )
        self._project_id = os.environ[self.PROJECT_ID]
        self._project_name = self._project_name()
        self._auth = auth.ApiKeyAuth(self._get_serving_api_key())
        self._verify = self._get_verify(hostname_verification, trust_store_path)
        self._session = requests.session()

        self._connected = True

    def _project_name(self):
        try:
            return os.environ[self.PROJECT_NAME]
        except KeyError:
            pass

        hops_user = self._project_user()
        hops_user_split = hops_user.split(
            "__"
        )  # project users have username project__user
        project = hops_user_split[0]
        return project

    def _project_user(self):
        try:
            hops_user = os.environ[self.HADOOP_USER_NAME]
        except KeyError:
            hops_user = os.environ[self.HDFS_USER]
        return hops_user

    def _get_ca_chain_path(self) -> str:
        return os.path.join("/tmp", self.PEM_CA_CHAIN)

    def _get_serving_api_key(self):
        """Retrieve serving API key from environment variable."""
        if self.SERVING_API_KEY not in os.environ:
            raise exceptions.InternalClientError("Serving API key not found")
        return os.environ[self.SERVING_API_KEY]

    def replace_public_host(self, url):
        """replace hostname to public hostname set in HOPSWORKS_PUBLIC_HOST"""
        ui_url = url._replace(netloc=os.environ[self.HOPSWORKS_PUBLIC_HOST])
        return ui_url

    def _is_external(self):
        return False



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Optional

import humps
from hopsworks_common import util
from hsml.predictor_state_condition import PredictorStateCondition


class PredictorState:
    """State of a predictor."""

    def __init__(
        self,
        available_predictor_instances: int,
        available_transformer_instances: Optional[int],
        hopsworks_inference_path: str,
        model_server_inference_path: str,
        internal_port: Optional[int],
        revision: Optional[int],
        deployed: Optional[bool],
        condition: Optional[PredictorStateCondition],
        status: str,
        **kwargs,
    ):
        self._available_predictor_instances = available_predictor_instances
        self._available_transformer_instances = available_transformer_instances
        self._hopsworks_inference_path = hopsworks_inference_path
        self._model_server_inference_path = model_server_inference_path
        self._internal_port = internal_port
        self._revision = revision
        self._deployed = deployed if deployed is not None else False
        self._condition = condition
        self._status = status

    def describe(self):
        """Print a description of the deployment state"""
        util.pretty_print(self)

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return PredictorState(*cls.extract_fields_from_json(json_decamelized))

    @classmethod
    def extract_fields_from_json(cls, json_decamelized):
        ai = util.extract_field_from_json(json_decamelized, "available_instances")
        ati = util.extract_field_from_json(
            json_decamelized, "available_transformer_instances"
        )
        hip = util.extract_field_from_json(json_decamelized, "hopsworks_inference_path")
        msip = util.extract_field_from_json(
            json_decamelized, "model_server_inference_path"
        )
        ipt = util.extract_field_from_json(json_decamelized, "internal_port")
        r = util.extract_field_from_json(json_decamelized, "revision")
        d = util.extract_field_from_json(json_decamelized, "deployed")
        c = util.extract_field_from_json(
            json_decamelized, "condition", as_instance_of=PredictorStateCondition
        )
        s = util.extract_field_from_json(json_decamelized, "status")

        return ai, ati, hip, msip, ipt, r, d, c, s

    def to_dict(self):
        json = {
            "availableInstances": self._available_predictor_instances,
            "hopsworksInferencePath": self._hopsworks_inference_path,
            "modelServerInferencePath": self._model_server_inference_path,
            "status": self._status,
        }

        if self._available_transformer_instances is not None:
            json["availableTransformerInstances"] = (
                self._available_transformer_instances
            )
        if self._internal_port is not None:
            json["internalPort"] = self._internal_port
        if self._revision is not None:
            json["revision"] = self._revision
        if self._deployed is not None:
            json["deployed"] = self._deployed
        if self._condition is not None:
            json = {**json, **self._condition.to_dict()}

        return json

    @property
    def available_predictor_instances(self):
        """Available predicotr instances."""
        return self._available_predictor_instances

    @property
    def available_transformer_instances(self):
        """Available transformer instances."""
        return self._available_transformer_instances

    @property
    def hopsworks_inference_path(self):
        """Inference path in the Hopsworks REST API."""
        return self._hopsworks_inference_path

    @property
    def model_server_inference_path(self):
        """Inference path in the model server"""
        return self._model_server_inference_path

    @property
    def internal_port(self):
        """Internal port for the predictor."""
        return self._internal_port

    @property
    def revision(self):
        """Last revision of the predictor."""
        return self._revision

    @property
    def deployed(self):
        """Whether the predictor is deployed or not."""
        return self._deployed

    @property
    def condition(self):
        """Condition of the current state of predictor."""
        return self._condition

    @property
    def status(self):
        """Status of the predictor."""
        return self._status

    def __repr__(self):
        return f"PredictorState(status: {self.status.capitalize()!r})"



import logging
import re

import pandas as pd
from hopsworks_common.core.constants import HAS_POLARS


logger = logging.getLogger(__name__)


class DataFrameValidator:
    # Base validator class

    @staticmethod
    def get_validator(df):
        """method to get the appropriate implementation of validator for the DataFrame type"""
        if isinstance(df, pd.DataFrame):
            return PandasValidator()

        if HAS_POLARS:
            import polars as pl

            if isinstance(df, pl.DataFrame):
                return PolarsValidator()

        try:
            from pyspark.sql import DataFrame as SparkDataFrame

            if isinstance(df, SparkDataFrame):
                return PySparkValidator()
        except ImportError:
            pass

        return None

    def _format_error_dict(self, errors):
        """Format error dictionary as a readable table string."""
        if not errors:
            return "{}"

        # Find maximum column name length for alignment
        col_width = max(len(col) for col in errors.keys()) + 2

        # Create table headers and separator
        result = ["", "Error details:"]
        result.append(f"{'Column':<{col_width}} | Error Description")
        result.append(f"{'-' * col_width}-+-{'-' * 50}")

        # Add each error row
        for col, desc in errors.items():
            result.append(f"{col:<{col_width}} | {desc}")

        return "\n".join(result)

    def _raise_validation_error(self, errors):
        """Raise a formatted validation error."""
        base_message = "One or more schema validation errors found for online ingestion. For details about the validation, refer to the documentation(https://docs.hopsworks.ai/latest/user_guides/fs/feature_group/data_types/#string-online-data-types)."
        formatted_errors = self._format_error_dict(errors)
        raise ValueError(f"{base_message}{formatted_errors}")

    def validate_schema(self, feature_group, df, df_features):
        """Common validation rules"""
        if feature_group.online_enabled is False:
            logger.warning("Feature group is not online enabled. Skipping validation")
            return df_features
        if feature_group._embedding_index is not None:
            logger.warning("Feature group is embedding type. Skipping validation")
            return df_features

        validator = self.get_validator(df)
        if validator is None:
            # If no validator is found for this type, skip validation and return df_features
            return df_features
        # If this is the base class and not a subclass instance being called directly,
        # delegate to the appropriate subclass
        if self.__class__ == DataFrameValidator:
            return validator.validate_schema(feature_group, df, df_features)
        errors = {}
        # Check if the primary key columns exist
        for pk in feature_group.primary_key:
            if pk not in df.columns:
                raise ValueError(
                    f"Primary key column {pk} is missing in input dataframe"
                )
        # Execute data type specific validation
        errors, column_lengths, is_pk_null, is_string_length_exceeded = (
            self._validate_df_specifics(feature_group, df)
        )

        # Handle errors
        if is_pk_null or (
            is_string_length_exceeded and (feature_group.id or feature_group.features)
        ):
            self._raise_validation_error(errors)
        elif is_string_length_exceeded:
            # If the feature group is not created and string lengths exceed default, adjust the string columns
            df_features = self.increase_string_columns(column_lengths, df_features)

        return df_features

    def _validate_df_specifics(self, feature_group, df):
        """To be implemented by subclasses"""
        raise NotImplementedError("Subclasses must implement this method")

    @staticmethod
    def get_feature_from_list(feature_name, features):
        for i_feature in features:
            if i_feature.name == feature_name:
                return i_feature

        return None

    @staticmethod
    def extract_numbers(input_string):
        # Define regular expression pattern for matching numbers
        pattern = r"\d+"
        # Use re.findall() to find all occurrences of the pattern in the input string
        return re.findall(pattern, input_string)

    def get_online_varchar_length(self, feature):
        # check of online_type is not null and starts with varchar
        if (
            feature
            and feature.online_type
            and feature.online_type.startswith("varchar")
        ):
            return int(self.extract_numbers(feature.online_type)[0])
        return None

    @staticmethod
    def increase_string_columns(column_lengths: dict, dataframe_features):
        def round_up_to_hundred(num):
            import math

            return math.ceil(num / 100) * 100

        # Adjusts the length of varchar columns
        for i_feature in dataframe_features:
            if i_feature.name in column_lengths:
                char_limit = round_up_to_hundred(column_lengths[i_feature.name])
                i_feature.online_type = f"varchar({char_limit})"
                logger.warning(
                    f"Maximum string length for column {i_feature.name} increased to {char_limit} in online table."
                )
        return dataframe_features


class PandasValidator(DataFrameValidator):
    @staticmethod
    def get_string_columns(df):
        if pd.__version__.startswith("2."):
            # For pandas 2+, use is_string_dtype api
            return [
                col
                for col in df.columns
                if pd.api.types.is_string_dtype(df[col].dropna())
            ]
        else:
            # For pandas 1.x,  is_string_dtype api is not compatible, so check each row if its a string
            string_cols = []
            for col in df.select_dtypes(include=["object", "string"]).columns:
                # Skip empty columns
                if df[col].count() == 0:
                    continue
                # Check if ALL non-null values are strings
                if df[col].dropna().map(lambda x: isinstance(x, str)).all():
                    string_cols.append(col)
            return string_cols

    # Pandas df specific validator
    def _validate_df_specifics(self, feature_group, df):
        errors = {}
        column_lengths = {}
        is_pk_null = False
        is_string_length_exceeded = False

        # Check for null values in primary key columns
        for pk in feature_group.primary_key:
            if df[pk].isnull().any():
                errors[pk] = f"Primary key column {pk} contains null values."
                is_pk_null = True

        # Check string lengths
        string_columns = self.get_string_columns(df)
        for col in string_columns:
            currentmax = df[col].str.len().max()
            col_max_len = (
                self.get_online_varchar_length(
                    self.get_feature_from_list(col, feature_group.features)
                )
                if feature_group.features
                else 100
            )

            if col_max_len is not None and currentmax > col_max_len:
                errors[col] = (
                    f"String length exceeded. Column {col} has string values longer than maximum column limit of {col_max_len} characters."
                )
                column_lengths[col] = currentmax
                is_string_length_exceeded = True

        return errors, column_lengths, is_pk_null, is_string_length_exceeded


class PolarsValidator(DataFrameValidator):
    # Polars df specific validator
    def _validate_df_specifics(self, feature_group, df):
        import polars as pl

        errors = {}
        column_lengths = {}
        is_pk_null = False
        is_string_length_exceeded = False

        # Check for null values in primary key columns
        for pk in feature_group.primary_key:
            if df[pk].is_null().any():
                errors[pk] = f"Primary key column {pk} contains null values."
                is_pk_null = True

        # Check string lengths
        for col in df.select(pl.col(pl.String)).columns:
            currentmax = df[col].str.len_chars().max()
            col_max_len = (
                self.get_online_varchar_length(
                    self.get_feature_from_list(col, feature_group.features)
                )
                if feature_group.features
                else 100
            )

            if col_max_len is not None and currentmax > col_max_len:
                errors[col] = (
                    f"String length exceeded. Column {col} has string values longer than maximum column limit of {col_max_len} characters."
                )
                column_lengths[col] = currentmax
                is_string_length_exceeded = True

        return errors, column_lengths, is_pk_null, is_string_length_exceeded


class PySparkValidator(DataFrameValidator):
    # PySpark-specific validator
    def _validate_df_specifics(self, feature_group, df):
        # Import PySpark SQL functions and types
        import pyspark.sql.functions as sf
        from pyspark.sql.types import StringType

        errors = {}
        column_lengths = {}
        is_pk_null = False
        is_string_length_exceeded = False

        # Check for null values in primary key columns
        for pk in feature_group.primary_key:
            if df.filter(df[pk].isNull()).count() > 0:
                errors[pk] = f"Primary key column {pk} contains null values."
                is_pk_null = True

        # Check string lengths for string columns
        for field in df.schema.fields:
            if isinstance(field.dataType, StringType):
                col = field.name
                # Compute max length - PySpark specific way
                currentmax_row = df.select(sf.max(sf.length(col))).collect()[0][0]
                currentmax = 0 if currentmax_row is None else currentmax_row

                col_max_len = (
                    self.get_online_varchar_length(
                        self.get_feature_from_list(col, feature_group.features)
                    )
                    if feature_group.features
                    else 100
                )

                if col_max_len is not None and currentmax > col_max_len:
                    errors[col] = (
                        f"String length exceeded. Column {col} has string values longer than maximum column limit of {col_max_len} characters."
                    )
                    column_lengths[col] = currentmax
                    is_string_length_exceeded = True

        return errors, column_lengths, is_pk_null, is_string_length_exceeded



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import Any, Dict

import humps
from hopsworks_common import util


class OnlineIngestionResult:
    """
    Metadata object used to provide Online Ingestion Batch Result information.

    This class encapsulates the result of a single batch operation during online ingestion,
    including the ingestion ID, status, and number of rows processed.
    """

    def __init__(
        self,
        online_ingestion_id: int = None,
        status: str = None,
        rows: int = None,
        **kwargs,
    ):
        """
        Initialize an OnlineIngestionResult object.

        # Arguments
            online_ingestion_id (int, optional): The unique identifier for the online ingestion batch.
            status (str, optional): The status of the ingestion batch (e.g., "UPSERTED", "FAILED").
            rows (int, optional): The number of rows processed in this batch.
        """
        self._online_ingestion_id = online_ingestion_id
        self._status = status
        self._rows = rows

    @classmethod
    def from_response_json(
        cls, json_dict: Dict[str, Any]
    ) -> "OnlineIngestionResult":
        """
        Create an OnlineIngestionResult object (or list of objects) from a JSON response.

        # Arguments
            json_dict (Dict[str, Any]): The JSON dictionary from the API response.

        # Returns
            OnlineIngestionResult or List[OnlineIngestionResult] or None: The created object(s), or None if input is None.
        """
        if json_dict is None:
            return None

        json_decamelized: dict = humps.decamelize(json_dict)

        if "count" not in json_decamelized:
            return cls(**json_decamelized)
        elif json_decamelized["count"] == 1:
            return cls(**json_decamelized["items"][0])
        elif json_decamelized["count"] > 1:
            return [cls(**item) for item in json_decamelized["items"]]
        else:
            return None

    def to_dict(self):
        """
        Convert the OnlineIngestionResult object to a dictionary.

        # Returns
            dict: Dictionary representation of the object.
        """
        return {
            "onlineIngestionId": self._online_ingestion_id,
            "status": self._status,
            "rows": self._rows,
        }

    def json(self):
        """
        Serialize the OnlineIngestionResult object to a JSON string.

        # Returns
            str: JSON string representation of the object.
        """
        return json.dumps(self, cls=util.Encoder)

    @property
    def online_ingestion_id(self) -> int:
        """
        Get the unique identifier for the online ingestion batch.

        # Returns
            int: The online ingestion batch ID.
        """
        return self._online_ingestion_id

    @property
    def status(self) -> str:
        """
        Get the status of the ingestion batch.

        # Returns
            str: The status of the batch (e.g., "UPSERTED", "FAILED").
        """
        return self._status

    @property
    def rows(self) -> int:
        """
        Get the number of rows processed in this batch.

        # Returns
            int: The number of rows.
        """
        return self._rows



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import warnings

from hopsworks_common import util, version
from hopsworks_common.connection import Connection


connection = Connection.connection

__version__ = version.__version__


def ml_formatwarning(message, category, filename, lineno, line=None):
    return "{}: {}\n".format(category.__name__, message)


warnings.formatwarning = ml_formatwarning
warnings.simplefilter("always", util.VersionWarning)

__all__ = ["connection"]



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hsml.constants import MODEL, PREDICTOR
from hsml.predictor import Predictor


class Predictor(Predictor):
    """Configuration for a predictor running with the vLLM backend"""

    def __init__(self, **kwargs):
        kwargs["model_framework"] = MODEL.FRAMEWORK_LLM
        kwargs["model_server"] = PREDICTOR.MODEL_SERVER_VLLM

        super().__init__(**kwargs)



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import unittest
from unittest import mock

from hsml.core import explicit_provenance


class TestExplicitProvenance(unittest.TestCase):
    def test_one_accessible_parent(self):
        artifact = {"id": 1}
        links = explicit_provenance.Links(accessible=[artifact])
        parent = explicit_provenance.Links.get_one_accessible_parent(links)
        self.assertEqual(artifact["id"], parent["id"])

    def test_one_accessible_parent_none(self):
        links = explicit_provenance.Links()
        with mock.patch.object(explicit_provenance._logger, "info") as mock_logger:
            parent = explicit_provenance.Links.get_one_accessible_parent(links)
            mock_logger.assert_called_once_with("There is no parent information")
            self.assertIsNone(parent)

    def test_one_accessible_parent_inaccessible(self):
        artifact = {"id": 1}
        links = explicit_provenance.Links(inaccessible=[artifact])
        with mock.patch.object(explicit_provenance._logger, "info") as mock_logger:
            parent = explicit_provenance.Links.get_one_accessible_parent(links)
            mock_logger.assert_called_once_with(
                "The parent is deleted or inaccessible. For more details get the full provenance from `_provenance` method"
            )
            self.assertIsNone(parent)

    def test_one_accessible_parent_deleted(self):
        artifact = {"id": 1}
        links = explicit_provenance.Links(deleted=[artifact])
        with mock.patch.object(explicit_provenance._logger, "info") as mock_logger:
            parent = explicit_provenance.Links.get_one_accessible_parent(links)
            mock_logger.assert_called_once_with(
                "The parent is deleted or inaccessible. For more details get the full provenance from `_provenance` method"
            )
            self.assertIsNone(parent)

    def test_one_accessible_parent_too_many(self):
        artifact1 = {"id": 1}
        artifact2 = {"id": 2}
        links = explicit_provenance.Links(accessible=[artifact1, artifact2])
        with self.assertRaises(Exception) as context:
            explicit_provenance.Links.get_one_accessible_parent(links)
            self.assertTrue(
                "Backend inconsistency - provenance returned more than one parent"
                in context.exception
            )

    def test_one_accessible_parent_should_not_be_artifact(self):
        artifact = explicit_provenance.Artifact(
            1, "test", 1, None, explicit_provenance.Artifact.MetaType.NOT_SUPPORTED
        )
        links = explicit_provenance.Links(accessible=[artifact])
        with self.assertRaises(Exception) as context:
            explicit_provenance.Links.get_one_accessible_parent(links)
            self.assertTrue(
                "The returned object is not a valid object. For more details get the full provenance from `_provenance` method"
                in context.exception
            )



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Dict, Mapping, Optional, Union

import humps


@dataclass
class FeatureTransformationStatistics:
    """
    Data class that contains all the statistics parameters that can be used for transformations inside a custom transformation function.
    """

    feature_name: str
    count: int = None
    # for any feature type
    completeness: Optional[float] = None
    num_non_null_values: Optional[int] = None
    num_null_values: Optional[int] = None
    approx_num_distinct_values: Optional[int] = None
    # for numerical features
    min: Optional[float] = None
    max: Optional[float] = None
    sum: Optional[float] = None
    mean: Optional[float] = None
    stddev: Optional[float] = None
    percentiles: Optional[Mapping[str, float]] = None
    # with exact uniqueness
    distinctness: Optional[float] = None
    entropy: Optional[float] = None
    uniqueness: Optional[float] = None
    exact_num_distinct_values: Optional[int] = None
    extended_statistics: Optional[Union[dict, str]] = None

    def __init__(
        self,
        feature_name: str,
        count: int = None,
        completeness: Optional[float] = None,
        num_non_null_values: Optional[int] = None,
        num_null_values: Optional[int] = None,
        approx_num_distinct_values: Optional[int] = None,
        min: Optional[float] = None,
        max: Optional[float] = None,
        sum: Optional[float] = None,
        mean: Optional[float] = None,
        stddev: Optional[float] = None,
        percentiles: Optional[Mapping[str, float]] = None,
        distinctness: Optional[float] = None,
        entropy: Optional[float] = None,
        uniqueness: Optional[float] = None,
        exact_num_distinct_values: Optional[int] = None,
        extended_statistics: Optional[Union[dict, str]] = None,
        **kwargs,
    ):
        self._feature_name = feature_name
        self._count = count
        self._completeness = completeness
        self._num_non_null_values = num_non_null_values
        self._num_null_values = num_null_values
        self._approx_num_distinct_values = approx_num_distinct_values
        self._min = min
        self._max = max
        self._sum = sum
        self._mean = mean
        self._stddev = stddev
        self._percentiles = percentiles
        self._distinctness = distinctness
        self._entropy = entropy
        self._uniqueness = uniqueness
        self._exact_num_distinct_values = exact_num_distinct_values
        extended_statistics = (
            extended_statistics
            if not isinstance(extended_statistics, str)
            else json.loads(extended_statistics)
        )
        if extended_statistics:
            self._correlations = extended_statistics.get("correlations", None)
            self._histogram = extended_statistics.get("histogram", None)
            self._kll = extended_statistics.get("kll", None)
            self._unique_values = extended_statistics.get("unique_values", None)

    @property
    def feature_name(self) -> str:
        """Name of the feature."""
        return self._feature_name

    @property
    def count(self) -> int:
        """Number of values."""
        return self._count

    @property
    def completeness(self) -> Optional[float]:
        """Fraction of non-null values in a column."""
        return self._completeness

    @property
    def num_non_null_values(self) -> Optional[int]:
        """Number of non-null values."""
        return self._num_non_null_values

    @property
    def num_null_values(self) -> Optional[int]:
        """Number of null values."""
        return self._num_null_values

    @property
    def approx_num_distinct_values(self) -> Optional[int]:
        """Approximate number of distinct values."""
        return self._approx_num_distinct_values

    @property
    def min(self) -> Optional[float]:
        """Minimum value."""
        return self._min

    @property
    def max(self) -> Optional[float]:
        """Maximum value."""
        return self._max

    @property
    def sum(self) -> Optional[float]:
        """Sum of all feature values."""
        return self._sum

    @property
    def mean(self) -> Optional[float]:
        """Mean value."""
        return self._mean

    @property
    def stddev(self) -> Optional[float]:
        """Standard deviation of the feature values."""
        return self._stddev

    @property
    def percentiles(self) -> Optional[Mapping[str, float]]:
        """Percentiles."""
        return self._percentiles

    @property
    def distinctness(self) -> Optional[float]:
        """Fraction of distinct values of a feature over the number of all its values. Distinct values occur at least once.

        !!! note "Example"
            $[a, a, b]$ contains two distinct values $a$ and $b$, so distinctness is $2/3$.
        """
        return self._distinctness

    @property
    def entropy(self) -> Optional[float]:
        """Entropy is a measure of the level of information contained in an event (feature value) when considering all possible events (all feature values).
        Entropy is estimated using observed value counts as the negative sum of (value_count/total_count) * log(value_count/total_count).

        !!! note "Example"
            $[a, b, b, c, c]$ has three distinct values with counts $[1, 2, 2]$.

            Entropy is then $(-1/5*log(1/5)-2/5*log(2/5)-2/5*log(2/5)) = 1.055$.
        """
        return self._entropy

    @property
    def uniqueness(self) -> Optional[float]:
        """Fraction of unique values over the number of all values of a column. Unique values occur exactly once.

        !!! note "Example"
            $[a, a, b]$ contains one unique value $b$, so uniqueness is $1/3$.
        """
        return self._uniqueness

    @property
    def exact_num_distinct_values(self) -> Optional[int]:
        """Exact number of distinct values."""
        return self._exact_num_distinct_values

    @property
    def correlations(self) -> Optional[dict]:
        """Correlations of feature values."""
        return self._correlations

    @property
    def histogram(self) -> Optional[dict]:
        """Histogram of feature values."""
        return self._histogram

    @property
    def kll(self) -> Optional[dict]:
        """KLL of feature values."""
        return self._kll

    @property
    def unique_values(self) -> Optional[dict]:
        """Number of Unique Values."""
        return self._unique_values

    @classmethod
    def from_response_json(
        cls: FeatureTransformationStatistics, json_dict: Dict[str, Any]
    ) -> FeatureTransformationStatistics:
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)


class TransformationStatistics:
    """
    Class that stores feature transformation statistics of all features that require training dataset statistics in a transformation function.

    All statistics for a feature is initially initialized with null values and will be populated with values when training dataset is created for the soe.

    # Arguments
        *features: `str`.
            The features for which training dataset statistics need to be computed.

    !!! example
        ```python
        # Defining transformation statistics
        transformation_statistics = TransformationStatistics("feature1", "feature2")

        # Accessing feature transformation statistics for a specific feature
        feature_transformation_statistics_feature1 = transformation_statistics.feature1
        ```
    """

    def __init__(self, *features: str):
        self._features = features
        self.__dict__.update(
            {feature: self.init_statistics(feature) for feature in features}
        )

    def init_statistics(self, feature_name: str) -> FeatureTransformationStatistics:
        return FeatureTransformationStatistics(feature_name=feature_name)

    def set_statistics(self, feature_name: str, statistics: Dict[str, Any]) -> None:
        self.__dict__[feature_name] = (
            FeatureTransformationStatistics.from_response_json(statistics)
        )

    def __repr__(self) -> str:
        return ",\n ".join([repr(self.__dict__[feature]) for feature in self._features])



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
import warnings

from hsfs import util


class DeltaStreamerJobConf:
    def __init__(self, options, spark_options, **kwargs):
        self._options = options
        self._spark_options = spark_options

    def to_dict(self):
        return {
            "writeOptions": self._options,
            "sparkJobConfiguration": self._spark_options,
        }

    def json(self):
        # no null values are allowed in the configuration
        null_options = {k: v for k, v in self._options.items() if v is None}
        if len(null_options) > 0:
            warnings.warn(
                f"None values are not allowed in Delta Streamer Job Configuration, dropping {null_options.keys()}.",
                stacklevel=2,
            )
            self._options = {k: v for k, v in self._options.items() if v is not None}
        return json.dumps(self, cls=util.Encoder)



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import (
    Any,
    Dict,
    Optional,
)

import humps
from hopsworks_common import util


class DataSource:
    """
    Metadata object used to provide Data source information for a feature group.
    """

    def __init__(
        self,
        query: Optional[str] = None,
        database: Optional[str] = None,
        group: Optional[str] = None,
        table: Optional[str] = None,
        path: Optional[str] = None,
        **kwargs,
    ):
        self._query = query
        self._database = database
        self._group = group
        self._table = table
        self._path = path

    @classmethod
    def from_response_json(
        cls, json_dict: Dict[str, Any]
    ) -> DataSource:
        if json_dict is None:
            return None

        json_decamelized: dict = humps.decamelize(json_dict)

        if "items" not in json_decamelized:
            return cls(**json_decamelized)
        else:
            return [
                cls(**item)
                for item in json_decamelized["items"]
            ]

    def to_dict(self):
        return {
            "query": self._query,
            "database": self._database,
            "group": self._group,
            "table": self._table,
            "path": self._path
        }

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    @property
    def query(self) -> Optional[str]:
        return self._query

    @query.setter
    def query(self, query: str) -> None:
        self._query = query

    @property
    def database(self) -> Optional[str]:
        return self._database

    @database.setter
    def database(self, database: str) -> None:
        self._database = database

    @property
    def group(self) -> Optional[str]:
        return self._group

    @group.setter
    def group(self, group: str) -> None:
        self._group = group

    @property
    def table(self) -> Optional[str]:
        return self._table

    @table.setter
    def table(self, table: str) -> None:
        self._table = table

    @property
    def path(self) -> Optional[str]:
        return self._path

    @path.setter
    def path(self, path: str) -> None:
        self._path = path



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Optional, Union

import humps
from hopsworks_common import client, util
from hopsworks_common.constants import RESOURCES, Default
from hsml.deployable_component import DeployableComponent
from hsml.resources import TransformerResources


class Transformer(DeployableComponent):
    """Metadata object representing a transformer to be used in a predictor."""

    def __init__(
        self,
        script_file: str,
        resources: Optional[Union[TransformerResources, dict, Default]] = None,  # base
        **kwargs,
    ):
        resources = (
            self._validate_resources(
                util.get_obj_from_json(resources, TransformerResources)
            )
            or self._get_default_resources()
        )
        if resources.num_instances is None:
            resources.num_instances = self._get_default_num_instances()

        super().__init__(script_file, resources)

    def describe(self):
        """Print a description of the transformer"""
        util.pretty_print(self)

    @classmethod
    def _validate_resources(cls, resources):
        if resources is not None:
            # ensure scale-to-zero for kserve deployments when required
            if resources.num_instances != 0 and client.is_scale_to_zero_required():
                raise ValueError(
                    "Scale-to-zero is required for KServe deployments in this cluster. Please, set the number of transformer instances to 0."
                )
        return resources

    @classmethod
    def _get_default_num_instances(cls):
        return (
            0  # enable scale-to-zero by default if required
            if client.is_scale_to_zero_required()
            else RESOURCES.MIN_NUM_INSTANCES
        )

    @classmethod
    def _get_default_resources(cls):
        return TransformerResources(cls._get_default_num_instances())

    @classmethod
    def from_json(cls, json_decamelized):
        sf, rc = cls.extract_fields_from_json(json_decamelized)
        return Transformer(sf, rc) if sf is not None else None

    @classmethod
    def extract_fields_from_json(cls, json_decamelized):
        sf = util.extract_field_from_json(
            json_decamelized, ["transformer", "script_file"]
        )
        rc = TransformerResources.from_json(json_decamelized)
        return sf, rc

    def update_from_response_json(self, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(*self.extract_fields_from_json(json_decamelized))
        return self

    def to_dict(self):
        return {"transformer": self._script_file, **self._resources.to_dict()}

    def __repr__(self):
        return f"Transformer({self._script_file!r})"



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
from typing import Optional

import humps
from hopsworks_common import util


class Alert:
    def __init__(
        self,
        id=None,
        status=None,
        alert_type=None,
        severity=None,
        receiver=None,
        created=None,
        **kwargs,
    ):
        self._id = id
        self._status = status
        self._alert_type = alert_type
        self._severity = severity
        self._receiver = receiver
        self._created = created

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if "count" in json_decamelized:
            if "items" in json_decamelized:
                return [cls(**receiver) for receiver in json_decamelized["items"]]
            else:
                return []
        else:
            return cls(**json_decamelized)

    @property
    def id(self) -> Optional[int]:
        """return the id of the alert"""
        return self._id

    @property
    def status(self) -> Optional[str]:
        """return the status of the alert"""
        return self._status

    @property
    def alert_type(self) -> Optional[str]:
        """return the type of the alert"""
        return self._alert_type

    @property
    def severity(self) -> Optional[str]:
        """return the severity of the alert"""
        return self._severity

    @property
    def receiver(self) -> Optional[str]:
        """return the receiver of the alert"""
        return self._receiver

    @property
    def created(self) -> Optional[str]:
        """return the creation time of the alert"""
        return self._created

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """return the alert as a dictionary"""
        return {
            "id": self._id,
            "status": self._status,
            "severity": self._severity,
            "receiver": self._receiver,
            "created": self._created,
        }

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"Alert({self._id!r}, {self._status!r}, {self._severity!r}, {self._receiver!r}, {self._created!r})"


class ProjectAlert(Alert):
    NOT_FOUND_ERROR_CODE = 150078

    def __init__(
        self,
        id=None,
        project_name=None,
        status=None,
        alert_type=None,
        severity=None,
        receiver=None,
        created=None,
        service=None,
        threshold=None,
        **kwargs,
    ):
        super().__init__(
            id=id,
            status=status,
            alert_type=alert_type,
            severity=severity,
            receiver=receiver,
            created=created,
            **kwargs,
        )
        self._project_name = project_name
        self._threshold = threshold
        self._service = service

    @property
    def project_name(self) -> Optional[str]:
        """return the name of the project"""
        return self._project_name

    @property
    def service(self) -> Optional[str]:
        """return the name of the service"""
        return self._service

    @property
    def threshold(self) -> Optional[str]:
        """return the threshold of the alert"""
        return self._threshold

    def to_dict(self):
        """return the alert as a dictionary"""
        return {
            "id": self._id,
            "project_name": self._project_name,
            "status": self._status,
            "severity": self._severity,
            "receiver": self._receiver,
            "created": self._created,
            "service": self._service,
            "threshold": self._threshold,
        }

    def __repr__(self):
        return f"ProjectAlert({self._id!r}, {self._project_name!r}, {self._status!r}, {self._severity!r}, {self._receiver!r}, {self._created!r}, {self._service!r}, {self._threshold!r})"


class JobAlert(Alert):
    NOT_FOUND_ERROR_CODE = 130034

    def __init__(
        self,
        id=None,
        job_name=None,
        status=None,
        alert_type=None,
        severity=None,
        receiver=None,
        created=None,
        threshold=None,
        **kwargs,
    ):
        super().__init__(
            id=id,
            status=status,
            alert_type=alert_type,
            severity=severity,
            receiver=receiver,
            created=created,
            **kwargs,
        )
        self._job_name = job_name
        self._threshold = threshold

    @property
    def job_name(self) -> Optional[str]:
        """return the name of the job"""
        return self._job_name

    @property
    def threshold(self) -> Optional[str]:
        """return the threshold of the alert"""
        return self._threshold

    def to_dict(self):
        """return the alert as a dictionary"""
        return {
            "id": self._id,
            "job_name": self._job_name,
            "status": self._status,
            "severity": self._severity,
            "receiver": self._receiver,
            "created": self._created,
            "threshold": self._threshold,
        }

    def __repr__(self):
        return f"JobAlert({self._id!r}, {self._job_name!r}, {self._status!r}, {self._severity!r}, {self._receiver!r}, {self._created!r}, {self._threshold!r})"


class FeatureGroupAlert(Alert):
    NOT_FOUND_ERROR_CODE = 270155

    def __init__(
        self,
        id=None,
        feature_store_name=None,
        feature_group_id=None,
        feature_group_name=None,
        status=None,
        alert_type=None,
        severity=None,
        receiver=None,
        created=None,
        **kwargs,
    ):
        super().__init__(
            id=id,
            status=status,
            alert_type=alert_type,
            severity=severity,
            receiver=receiver,
            created=created,
            **kwargs,
        )
        self._feature_store_name = feature_store_name
        self._feature_group_id = feature_group_id
        self._feature_group_name = feature_group_name

    @property
    def feature_store_name(self) -> Optional[str]:
        """return the name of the feature store"""
        return self._feature_store_name

    @property
    def feature_group_id(self) -> Optional[str]:
        """return the id of the feature group"""
        return self._feature_group_id

    @property
    def feature_group_name(self) -> Optional[str]:
        """return the name of the feature group"""
        return self._feature_group_name

    def to_dict(self):
        """return the alert as a dictionary"""
        return {
            "id": self._id,
            "feature_store_name": self._feature_store_name,
            "feature_group_id": self._feature_group_id,
            "feature_group_name": self._feature_group_name,
            "status": self._status,
            "severity": self._severity,
            "receiver": self._receiver,
            "created": self._created,
        }


    def __repr__(self):
        return f"FeatureGroupAlert({self._id!r}, {self._feature_store_name!r}, {self._feature_group_id!r}, {self._feature_group_name!r}, {self._status!r}, {self._severity!r}, {self._receiver!r}, {self._created!r})"


class FeatureViewAlert(Alert):
    NOT_FOUND_ERROR_CODE = 270155

    def __init__(
        self,
        id=None,
        feature_store_name=None,
        feature_view_id=None,
        feature_view_name=None,
        feature_view_version=None,
        status=None,
        alert_type=None,
        severity=None,
        receiver=None,
        created=None,
        **kwargs,
    ):
        super().__init__(
            id=id,
            status=status,
            alert_type=alert_type,
            severity=severity,
            receiver=receiver,
            created=created,
            **kwargs,
        )
        self._feature_store_name = feature_store_name
        self._feature_view_id = feature_view_id
        self._feature_view_name = feature_view_name
        self._feature_view_version = feature_view_version

    @property
    def feature_store_name(self) -> Optional[str]:
        """return the name of the feature store"""
        return self._feature_store_name

    @property
    def feature_view_id(self) -> Optional[str]:
        """return the id of the feature view"""
        return self._feature_view_id

    @property
    def feature_view_name(self) -> Optional[str]:
        """return the name of the feature view"""
        return self._feature_view_name

    @property
    def feature_view_version(self) -> Optional[str]:
        """return the version of the feature view"""
        return self._feature_view_version

    def to_dict(self):
        """return the alert as a dictionary"""
        return {
            "id": self._id,
            "feature_store_name": self._feature_store_name,
            "feature_view_id": self._feature_view_id,
            "feature_view_name": self._feature_view_name,
            "feature_view_version": self._feature_view_version,
            "status": self._status,
            "severity": self._severity,
            "receiver": self._receiver,
            "created": self._created,
        }

    def __repr__(self):
        return f"FeatureViewAlert({self._id!r}, {self._feature_store_name!r}, {self._feature_view_id!r}, {self._feature_view_name!r}, {self._feature_view_version!r}, {self._status!r}, {self._severity!r}, {self._receiver!r}, {self._created!r})"



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Set, TypeVar, Union

import pandas as pd
from hsfs import feature_view, statistics, training_dataset, transformation_function
from hsfs.core import transformation_function_api


if TYPE_CHECKING:
    import polars as pl


class TransformationFunctionEngine:
    BUILTIN_FN_NAMES = [
        "min_max_scaler",
        "standard_scaler",
        "robust_scaler",
        "label_encoder",
    ]
    AMBIGUOUS_FEATURE_ERROR = (
        "Provided feature '{}' in transformation functions is ambiguous and exists in more than one feature groups."
        "You can provide the feature with the prefix that was specified in the join."
    )
    FEATURE_NOT_EXIST_ERROR = "Provided feature '{}' in transformation functions do not exist in any of the feature groups."

    def __init__(self, feature_store_id: int):
        self._feature_store_id = feature_store_id
        self._transformation_function_api: transformation_function_api.TransformationFunctionApi = transformation_function_api.TransformationFunctionApi(
            feature_store_id
        )

    def save(
        self, transformation_fn_instance: transformation_function.TransformationFunction
    ) -> transformation_function.TransformationFunction:
        """
        Save a transformation function into the feature store.

        # Argument
            transformation_fn_instance `transformation_function.TransformationFunction`: The transformation function to be saved into the feature store.
        """
        self._transformation_function_api.register_transformation_fn(
            transformation_fn_instance
        )

    def get_transformation_fn(
        self, name: str, version: Optional[int] = None
    ) -> Union[
        transformation_function.TransformationFunction,
        List[transformation_function.TransformationFunction],
    ]:
        """
        Retrieve a transformation function from the feature store.

        If only the name of the transformation function is provided then all the versions of the transformation functions are returned as a list.
        If both name and version are not provided then all transformation functions saved in the feature view is returned.

        # Argument
            name ` Optional[str]`: The name of the transformation function to be retrieved.
            version `Optional[int]`: The version of the transformation function to be retrieved.
        # Returns
            `Union[transformation_function.TransformationFunction, List[transformation_function.TransformationFunction]]` : A transformation function if name and version is provided. A list of transformation functions if only name is provided.
        """

        transformation_fn_instances = (
            self._transformation_function_api.get_transformation_fn(name, version)
        )
        return transformation_fn_instances

    def get_transformation_fns(
        self,
    ) -> List[transformation_function.TransformationFunction]:
        """
        Get all the transformation functions in the feature store

        # Returns
            `List[transformation_function.TransformationFunction]` : A list of transformation functions.
        """
        transformation_fn_instances = (
            self._transformation_function_api.get_transformation_fn(
                name=None, version=None
            )
        )
        transformation_fns = []
        for (
            transformation_fn_instance
        ) in transformation_fn_instances:  # todo what is the point of this?
            transformation_fns.append(transformation_fn_instance)
        return transformation_fns

    def delete(
        self,
        transformation_function_instance: transformation_function.TransformationFunction,
    ) -> None:
        """
        Delete a transformation function from the feature store.

        # Arguments
            transformation_function_instance `transformation_function.TransformationFunction`: The transformation function to be removed from the feature store.
        """
        self._transformation_function_api.delete(transformation_function_instance)

    @staticmethod
    def compute_transformation_fn_statistics(
        training_dataset_obj: training_dataset.TrainingDataset,
        statistics_features: List[str],
        label_encoder_features: List[str],
        feature_dataframe: Union[
            pd.DataFrame, pl.DataFrame, TypeVar("pyspark.sql.DataFrame")
        ],
        feature_view_obj: feature_view.FeatureView,
    ) -> statistics.Statistics:
        """
        Compute the statistics required for a training dataset object.

        # Arguments
            training_dataset_obj `TrainingDataset`: The training dataset for which the statistics is to be computed.
            statistics_features `List[str]`: The list of features for which the statistics should be computed.
            label_encoder_features `List[str]`: Features used for label encoding.
            feature_dataframe `Union[pd.DataFrame, pl.DataFrame, ps.DataFrame]`: The dataframe that contains the data for which the statistics must be computed.
            feature_view_obj `FeatureView`: The feature view in which the training data is being created.
        # Returns
            `Statistics` : The statistics object that contains the statistics for each features.
        """
        return training_dataset_obj._statistics_engine.compute_transformation_fn_statistics(
            td_metadata_instance=training_dataset_obj,
            columns=statistics_features,
            label_encoder_features=label_encoder_features,  # label encoded features only
            feature_dataframe=feature_dataframe,
            feature_view_obj=feature_view_obj,
        )

    @staticmethod
    def get_ready_to_use_transformation_fns(
        feature_view: feature_view.FeatureView,
        training_dataset_version: Optional[int] = None,
    ) -> List[transformation_function.TransformationFunction]:
        """
        Function that updates statistics required for all transformation functions in the feature view based on training dataset version.

        # Arguments
            feature_view `FeatureView`: The feature view in which the training data is being created.
            training_dataset_version `TrainingDataset`: The training version used to update the statistics used in the transformation functions.
        # Returns
            `List[transformation_function.TransformationFunction]` : List of transformation functions.
        """
        # check if transformation functions require statistics
        is_stat_required = any(
            [
                tf.hopsworks_udf.statistics_required
                for tf in feature_view.transformation_functions
            ]
        )
        if not is_stat_required:
            td_tffn_stats = None
        else:
            # if there are any transformation functions that require statistics get related statistics and
            # populate with relevant arguments
            # there should be only one statistics object with before_transformation=true
            if training_dataset_version is None:
                raise ValueError(
                    "Training data version is required for transformation. Call `feature_view.init_serving(version)` "
                    "or `feature_view.init_batch_scoring(version)` to pass the training dataset version."
                    "Training data can be created by `feature_view.create_training_data` or `feature_view.training_data`."
                )
            td_tffn_stats = feature_view._statistics_engine.get(
                feature_view,
                before_transformation=True,
                training_dataset_version=training_dataset_version,
            )

        if is_stat_required and td_tffn_stats is None:
            raise ValueError(
                "No statistics available for initializing transformation functions."
                + "Training data can be created by `feature_view.create_training_data` or `feature_view.training_data`."
            )

        if is_stat_required:
            for transformation_function in feature_view.transformation_functions:
                transformation_function.transformation_statistics = (
                    td_tffn_stats.feature_descriptive_statistics
                )
        return feature_view.transformation_functions

    @staticmethod
    def compute_and_set_feature_statistics(
        training_dataset: training_dataset.TrainingDataset,
        feature_view_obj: feature_view.FeatureView,
        dataset: Union[
            Dict[
                str, Union[pd.DataFrame, pl.DataFrame, TypeVar("pyspark.sql.DataFrame")]
            ],
            Union[pd.DataFrame, pl.DataFrame, TypeVar("pyspark.sql.DataFrame")],
        ],
    ) -> None:
        """
        Function that computes and sets the statistics required for the UDF used for transformation.

        The function assigns the statistics computed to hopsworks UDF object so that the statistics can be used when UDF is executed.

        # Argument
            training_dataset_obj `TrainingDataset`: The training dataset for which the statistics is to be computed.
            feature_view `FeatureView`: The feature view in which the training data is being created.
            dataset `Union[Dict[str,  Union[pd.DataFrame, pl.DataFrame, ps.DataFrame]],  Union[pd.DataFrame, pl.DataFrame, ps.DataFrame]]`: A dataframe that conqtains the training data or a dictionary that contains both the training and test data.
        """
        statistics_features: Set[str] = set()
        label_encoder_features: Set[str] = set()

        # Finding the features for which statistics is required
        for tf in feature_view_obj.transformation_functions:
            statistics_features.update(tf.hopsworks_udf.statistics_features)
            if (
                tf.hopsworks_udf.function_name == "label_encoder"
                or tf.hopsworks_udf.function_name == "one_hot_encoder"
            ):
                label_encoder_features.update(tf.hopsworks_udf.statistics_features)
        if statistics_features:
            # compute statistics on training data
            if training_dataset.splits:
                # compute statistics before transformations are applied
                stats = (
                    TransformationFunctionEngine.compute_transformation_fn_statistics(
                        training_dataset,
                        list(statistics_features),
                        list(label_encoder_features),
                        dataset.get(training_dataset.train_split),
                        feature_view_obj,
                    )
                )
            else:
                stats = (
                    TransformationFunctionEngine.compute_transformation_fn_statistics(
                        training_dataset,
                        list(statistics_features),
                        list(label_encoder_features),
                        dataset,
                        feature_view_obj,
                    )
                )

            # Set statistics computed in the hopsworks UDF
            for tf in feature_view_obj.transformation_functions:
                tf.transformation_statistics = stats.feature_descriptive_statistics

    @staticmethod
    def get_and_set_feature_statistics(
        training_dataset: training_dataset.TrainingDataset,
        feature_view_obj: feature_view.FeatureView,
        training_dataset_version: int = None,
    ) -> None:
        """
        Function that gets the transformation statistics computed while creating the training dataset from the backend and assigns it to the hopsworks UDF object.

        The function assigns the statistics computed to hopsworks UDF object so that the statistics can be used when UDF is executed.

        # Argument
            training_dataset_obj `TrainingDataset`: The training dataset for which the statistics is to be computed.
            feature_view `FeatureView`: The feature view in which the training data is being created.
            training_dataset_version `int`: The version of the training dataset for which the statistics is to be retrieved.

        # Raises
            `ValueError` : If the statistics are not present in the backend.
        """

        is_stat_required = any(
            [
                tf.hopsworks_udf.statistics_required
                for tf in feature_view_obj.transformation_functions
            ]
        )

        if is_stat_required:
            td_tffn_stats = training_dataset._statistics_engine.get(
                feature_view_obj,
                before_transformation=True,
                training_dataset_version=training_dataset_version,
            )

            if td_tffn_stats is None:
                raise ValueError(
                    "No statistics available for initializing transformation functions."
                )

            for tf in feature_view_obj.transformation_functions:
                tf.transformation_statistics = (
                    td_tffn_stats.feature_descriptive_statistics
                )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import json
from typing import Any, Dict, List, Optional, Union

from hopsworks_common import (
    client,
    decorators,
    execution,
    job,
    job_schedule,
    usage,
    util,
)
from hopsworks_common.core import (
    ingestion_job_conf,
    job_configuration,
)


class JobApi:
    @usage.method_logger
    def create_job(self, name: str, config: dict) -> job.Job:
        """Create a new job or update an existing one.

        ```python

        import hopsworks

        project = hopsworks.login()

        job_api = project.get_job_api()

        spark_config = job_api.get_configuration("PYSPARK")

        spark_config['appPath'] = "/Resources/my_app.py"

        job = job_api.create_job("my_spark_job", spark_config)

        ```
        # Arguments
            name: Name of the job.
            config: Configuration of the job.
        # Returns
            `Job`: The Job object
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()

        config = util.validate_job_conf(config, _client._project_name)

        path_params = ["project", _client._project_id, "jobs", name]

        headers = {"content-type": "application/json"}
        created_job = job.Job.from_response_json(
            _client._send_request(
                "PUT", path_params, headers=headers, data=json.dumps(config)
            )
        )
        print(f"Job created successfully, explore it at {created_job.get_url()}")
        return created_job

    @usage.method_logger
    @decorators.catch_not_found("hopsworks_common.job.Job", fallback_return=None)
    def get_job(self, name: str) -> Optional[job.Job]:
        """Get a job.

        # Arguments
            name: Name of the job.
        # Returns
            `Job`: The Job object or `None` if it does not exist.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "jobs",
            name,
        ]
        query_params = {"expand": ["creator"]}
        return job.Job.from_response_json(
            _client._send_request("GET", path_params, query_params=query_params)
        )

    @usage.method_logger
    def get_jobs(self) -> List[job.Job]:
        """Get all jobs.

        # Returns
            `List[Job]`: List of Job objects
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "jobs",
        ]
        query_params = {"expand": ["creator"]}
        return job.Job.from_response_json(
            _client._send_request("GET", path_params, query_params=query_params)
        )

    @usage.method_logger
    def exists(self, name: str):
        """Check if a job exists.

        # Arguments
            name: Name of the job.
        # Returns
            `bool`: True if the job exists, otherwise False
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        job = self.get_job(name)
        return job is not None

    @usage.method_logger
    def get_configuration(self, type: str):
        """Get configuration for the specific job type.

        # Arguments
            type: Type of the job. Currently, supported types include: SPARK, PYSPARK, PYTHON, DOCKER, FLINK.
        # Returns
            `dict`: Default job configuration
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "jobs",
            type.lower(),
            "configuration",
        ]

        headers = {"content-type": "application/json"}
        return _client._send_request("GET", path_params, headers=headers)

    def _delete(self, job):
        """Delete the job and all executions.
        :param job: metadata object of job to delete
        :type job: Job
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "jobs",
            str(job.name),
        ]
        _client._send_request("DELETE", path_params)

    def _update_job(self, name: str, config: dict):
        """Update the job.
        :param name: name of the job
        :type name: str
        :param config: new job configuration
        :type config: dict
        :return: The updated Job object
        :rtype: Job
        """
        _client = client.get_instance()

        config = util.validate_job_conf(config, _client._project_name)

        path_params = ["project", _client._project_id, "jobs", name]

        headers = {"content-type": "application/json"}
        return job.Job.from_response_json(
            _client._send_request(
                "PUT", path_params, headers=headers, data=json.dumps(config)
            )
        )

    def _schedule_job(self, name, schedule_config):
        """Attach the `schedule_config` to the job with the given `name`."""
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", name, "schedule", "v2"]
        headers = {"content-type": "application/json"}
        method = "PUT" if schedule_config["id"] else "POST"

        return job_schedule.JobSchedule.from_response_json(
            _client._send_request(
                method, path_params, headers=headers, data=json.dumps(schedule_config)
            )
        )

    def _delete_schedule_job(self, name):
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", name, "schedule", "v2"]

        return _client._send_request(
            "DELETE",
            path_params,
        )

    @usage.method_logger
    def create(
        self,
        name: str,
        job_conf: Union[
            job_configuration.JobConfiguration, ingestion_job_conf.IngestionJobConf
        ],
    ) -> job.Job:
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", name]

        headers = {"content-type": "application/json"}
        return job.Job.from_response_json(
            _client._send_request(
                "PUT", path_params, headers=headers, data=job_conf.json()
            )
        )

    @usage.method_logger
    def launch(self, name: str, args: str = None) -> None:
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", name, "executions"]

        _client._send_request("POST", path_params, data=args)

    @usage.method_logger
    def get(self, name: str) -> job.Job:
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", name]

        return job.Job.from_response_json(_client._send_request("GET", path_params))

    @usage.method_logger
    def last_execution(self, job: job.Job) -> execution.Execution:
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", job.name, "executions"]

        query_params = {"limit": 1, "sort_by": "submissiontime:desc"}

        headers = {"content-type": "application/json"}
        return execution.Execution.from_response_json(
            _client._send_request(
                "GET", path_params, headers=headers, query_params=query_params
            ),
            job=job,
        )

    @usage.method_logger
    def create_or_update_schedule_job(
        self, name: str, schedule_config: Dict[str, Any]
    ) -> job_schedule.JobSchedule:
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", name, "schedule", "v2"]
        headers = {"content-type": "application/json"}
        method = "PUT" if schedule_config["id"] else "POST"

        return job_schedule.JobSchedule.from_response_json(
            _client._send_request(
                method, path_params, headers=headers, data=json.dumps(schedule_config)
            )
        )

    @usage.method_logger
    def delete_schedule_job(self, name: str) -> None:
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", name, "schedule", "v2"]

        return _client._send_request(
            "DELETE",
            path_params,
        )



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks_common.util import (
    FEATURE_STORE_NAME_SUFFIX,
    VALID_EMBEDDING_TYPE,
    Encoder,
    FeatureGroupWarning,
    JobWarning,
    StatisticsWarning,
    StorageWarning,
    ValidationWarning,
    VersionWarning,
    _loading_animation,
    append_feature_store_suffix,
    autofix_feature_name,
    check_timestamp_format_from_date_string,
    convert_event_time_to_timestamp,
    convert_git_status_to_files,
    convert_to_abs,
    feature_group_name,
    get_dataset_type,
    get_delta_datestr_from_timestamp,
    get_feature_group_url,
    get_hostname_replaced_url,
    get_hudi_datestr_from_timestamp,
    get_job_url,
    get_timestamp_from_date_string,
    is_interactive,
    is_runtime_notebook,
    run_with_loading_animation,
    strip_feature_store_suffix,
    validate_embedding_feature_type,
    validate_job_conf,
    verify_attribute_key_names,
)


__all__ = [
    "FEATURE_STORE_NAME_SUFFIX",
    "VALID_EMBEDDING_TYPE",
    "Encoder",
    "FeatureGroupWarning",
    "JobWarning",
    "StatisticsWarning",
    "StorageWarning",
    "ValidationWarning",
    "VersionWarning",
    "_loading_animation",
    "append_feature_store_suffix",
    "autofix_feature_name",
    "check_timestamp_format_from_date_string",
    "convert_event_time_to_timestamp",
    "convert_git_status_to_files",
    "convert_to_abs",
    "feature_group_name",
    "get_dataset_type",
    "get_delta_datestr_from_timestamp",
    "get_feature_group_url",
    "get_hostname_replaced_url",
    "get_hudi_datestr_from_timestamp",
    "get_job_url",
    "get_timestamp_from_date_string",
    "is_interactive",
    "is_runtime_notebook",
    "run_with_loading_animation",
    "strip_feature_store_suffix",
    "validate_embedding_feature_type",
    "validate_job_conf",
    "verify_attribute_key_names",
]



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks_common.client.istio.utils.infer_type import (
    GRPC_CONTENT_DATATYPE_MAPPINGS,
    InferenceServerException,
    InferInput,
    InferOutput,
    InferRequest,
    InferResponse,
    get_content,
    raise_error,
    serialize_byte_tensor,
)


__all__ = [
    "GRPC_CONTENT_DATATYPE_MAPPINGS",
    "InferenceServerException",
    "InferInput",
    "InferOutput",
    "InferRequest",
    "InferResponse",
    "get_content",
    "raise_error",
    "serialize_byte_tensor",
]



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import os


class HdfsApi:
    def __init__(self):
        import fsspec.implementations.arrow as pfs

        host, port = os.environ["LIBHDFS_DEFAULT_FS"].split(":")

        self._hopsfs = pfs.HadoopFileSystem(
            host=host,
            port=int(port),
            user=os.environ["LIBHDFS_DEFAULT_USER"],
        )

    DEFAULT_BUFFER_SIZE = 0

    def upload(
        self,
        local_path: str,
        upload_path: str,
        overwrite: bool = False,
        buffer_size: int = DEFAULT_BUFFER_SIZE,
    ):
        """Upload file/directory to the Hopsworks filesystem.
        :param local_path: local path to file to upload
        :type local_path: str
        :param upload_path: path to directory where to upload the file in Hopsworks filesystem
        :type upload_path: str
        :param overwrite: overwrite file if exists
        :type overwrite: bool
        :param buffer_size: size of the temporary read and write buffer. Defaults to 0.
        :type buffer_size: int
        """
        # local path could be absolute or relative,
        if not os.path.isabs(local_path) and os.path.exists(
            os.path.join(os.getcwd(), local_path)
        ):
            local_path = os.path.join(os.getcwd(), local_path)

        _, file_name = os.path.split(local_path)

        destination_path = upload_path + "/" + file_name

        if self._hopsfs.exists(destination_path):
            if overwrite:
                self._hopsfs.rm(destination_path, recursive=True)
            else:
                raise Exception(
                    "{} already exists, set overwrite=True to overwrite it".format(
                        local_path
                    )
                )

        self._hopsfs.upload(
            lpath=local_path,
            rpath=destination_path,
            recursive=True,
            buffer_size=buffer_size,
        )

        return upload_path + "/" + os.path.basename(local_path)

    def download(self, path, local_path, buffer_size=DEFAULT_BUFFER_SIZE):
        """Download file/directory on a path in datasets.
        :param path: path to download
        :type path: str
        :param local_path: path to download in datasets
        :type local_path: str
        :param buffer_size: size of the temporary read and write buffer. Defaults to 0.
        :type buffer_size: int
        """

        self._hopsfs.download(path, local_path, recursive=True, buffer_size=buffer_size)



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import (
    Any,
    Dict,
    Optional,
)

import humps
from hsfs import feature


class DataSourceData:
    """
    Metadata object used to provide Data source data information for a feature group.
    """

    def __init__(
        self,
        limit: Optional[int] = None,
        features: Optional[list[feature.Feature]] = None,
        preview: Optional[list[dict]] = None,
        **kwargs,
    ):
        self._limit = limit
        self._features = features
        self._preview = preview

    @classmethod
    def from_response_json(
        cls, json_dict: Dict[str, Any]
    ) -> DataSourceData:
        if json_dict is None:
            return None

        json_decamelized: dict = humps.decamelize(json_dict)

        return cls(**json_decamelized)

    @property
    def limit(self) -> Optional[int]:
        return self._limit

    @property
    def features(self) -> Optional[list[feature.Feature]]:
        return self._features

    @property
    def preview(self) -> Optional[list[dict]]:
        return self._preview



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import humps
from hsml import tag


class TestTag:
    # from response json

    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["tag"]["get"]["response"]
        json_camelized = humps.camelize(json)

        # Act
        t_list = tag.Tag.from_response_json(json_camelized)

        # Assert
        assert len(t_list) == 1
        t = t_list[0]
        assert t.name == "test_name"
        assert t.value == "test_value"

    def test_from_response_json_empty(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["tag"]["get_empty"]["response"]
        json_camelized = humps.camelize(json)

        # Act
        t_list = tag.Tag.from_response_json(json_camelized)

        # Assert
        assert len(t_list) == 0

    # constructor

    def test_constructor(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["tag"]["get"]["response"]["items"][0]
        tag_name = json.pop("name")
        tag_value = json.pop("value")

        # Act
        t = tag.Tag(name=tag_name, value=tag_value, **json)

        # Assert
        assert t.name == "test_name"
        assert t.value == "test_value"



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import List, Union

from hopsworks_common import client
from hsfs.core.variable_api import VariableApi
from hsfs.validation_report import ValidationReport


class ValidationReportApi:
    def __init__(self, feature_store_id: int, feature_group_id: int):
        """Validation Report endpoints for the featuregroup resource.

        :param feature_store_id: id of the respective featurestore
        :type feature_store_id: int
        :param feature_group_id: id of the respective featuregroup
        :type feature_group_id: int
        """
        self._feature_store_id = feature_store_id
        self._feature_group_id = feature_group_id
        self._variable_api = VariableApi()

    def create(self, validation_report: ValidationReport) -> ValidationReport:
        """Create an validation report attached to a featuregroup.

        :param validation_report: validation report object to be created for a featuregroup
        :type validation_report: `ValidationReport`

        :return: persisted validation report
        :rtype: `ValidationReport`
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "validationreport",
        ]

        major, minor = self._variable_api.parse_major_and_minor(
            self._variable_api.get_version("hopsworks")
        )
        if major == "3" and minor == "0":
            # You must bypass the setter otherwise it ends up as "UNKNOWN"
            validation_report._ingestion_result = None

        headers = {"content-type": "application/json"}
        payload = validation_report.json()

        return ValidationReport.from_response_json(
            _client._send_request("PUT", path_params, headers=headers, data=payload)
        )

    def delete(self, validation_report_id: int) -> None:
        """Delete the validation report attached to a featuregroup."""
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "validationreport",
            validation_report_id,
        ]

        _client._send_request("DELETE", path_params)

    def get_last(self) -> ValidationReport:
        """Gets the latest Validation Report of a featuregroup.

        :return: latest validation report
        :rtype: `ValidationReport`
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "validationreport",
        ]
        headers = {"content-type": "application/json"}
        query_params = {
            "sort_by": "validation_time:desc",
            "offset": 0,
            "limit": 1,
            "fields": "content",
        }

        return ValidationReport.from_response_json(
            _client._send_request("GET", path_params, query_params, headers=headers)
        )

    def get_all(self) -> Union[List[ValidationReport], ValidationReport]:
        """Get the validation report attached to a featuregroup.

        :return: validation report
        :rtype: Union[List[ValidationReport], ValidationReport]
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "validationreport",
        ]
        headers = {"content-type": "application/json"}
        query_params = {
            "sort_by": "validation_time:desc",
            "offset": 0,
            "fields": "content",
        }

        return ValidationReport.from_response_json(
            _client._send_request("GET", path_params, query_params, headers=headers)
        )



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from datetime import date, datetime, time

import pandas as pd
import pytest
from hsfs.client.exceptions import FeatureStoreException
from hsfs.hopsworks_udf import (
    HopsworksUdf,
    TransformationFeature,
    UDFExecutionMode,
    udf,
)


class TestHopsworksUdf:
    def test_udf_creation_invalid_execution_mode(self):
        with pytest.raises(FeatureStoreException) as exception:

            @udf(return_type=int, mode="invalid")
            def test(feature):
                pass

        assert (
            str(exception.value)
            == "Ivalid execution mode `invalid` for UDF. Please use `default`, `python` or `pandas` instead."
        )

    def test_udf_creation_default_execution_mode(self):
        @udf(return_type=int)
        def test(feature):
            pass

        assert test.execution_mode == UDFExecutionMode.DEFAULT

        @udf(return_type=int, mode="default")
        def test1(feature):
            pass

        assert test1.execution_mode == UDFExecutionMode.DEFAULT

    def test_udf_creation_python_execution_mode(self):
        @udf(return_type=int, mode="python")
        def test(feature):
            pass

        assert test.execution_mode == UDFExecutionMode.PYTHON

    def test_udf_creation_pandas_execution_mode(self):
        @udf(return_type=int, mode="pandas")
        def test(feature):
            pass

        assert test.execution_mode == UDFExecutionMode.PANDAS

    def test_validate_and_convert_output_types_one_elements(self):
        assert HopsworksUdf._validate_and_convert_output_types([int]) == ["bigint"]

        assert HopsworksUdf._validate_and_convert_output_types([float]) == ["double"]

        assert HopsworksUdf._validate_and_convert_output_types([str]) == ["string"]

        assert HopsworksUdf._validate_and_convert_output_types([bool]) == ["boolean"]

        assert HopsworksUdf._validate_and_convert_output_types([datetime]) == [
            "timestamp"
        ]

        assert HopsworksUdf._validate_and_convert_output_types([time]) == ["timestamp"]

        assert HopsworksUdf._validate_and_convert_output_types([date]) == ["date"]

        with pytest.raises(FeatureStoreException) as exception:
            HopsworksUdf._validate_and_convert_output_types([pd.DatetimeTZDtype])

        assert (
            str(exception.value)
            == f"Output type {pd.DatetimeTZDtype} is not supported. Please refer to the documentation to get more information on the supported types."
        )

    def test_validate_and_convert_output_types_multiple_types(self):
        assert HopsworksUdf._validate_and_convert_output_types(
            [int, float, str, bool, datetime, date, time]
        ) == ["bigint", "double", "string", "boolean", "timestamp", "date", "timestamp"]

        assert HopsworksUdf._validate_and_convert_output_types(
            ["bigint", "double", "string", "boolean", "timestamp", "date"]
        ) == ["bigint", "double", "string", "boolean", "timestamp", "date"]

        with pytest.raises(FeatureStoreException) as exception:
            HopsworksUdf._validate_and_convert_output_types([pd.DatetimeTZDtype])

        assert (
            str(exception.value)
            == f"Output type {pd.DatetimeTZDtype} is not supported. Please refer to the documentation to get more information on the supported types."
        )

    def test_validate_and_convert_output_types_invalid_types(self):
        with pytest.raises(FeatureStoreException) as exception:
            HopsworksUdf._validate_and_convert_output_types([pd.DatetimeTZDtype])

        assert (
            str(exception.value)
            == f"Output type {pd.DatetimeTZDtype} is not supported. Please refer to the documentation to get more information on the supported types."
        )

        with pytest.raises(FeatureStoreException) as exception:
            HopsworksUdf._validate_and_convert_output_types([int, pd.DatetimeTZDtype])

        assert (
            str(exception.value)
            == f"Output type {pd.DatetimeTZDtype} is not supported. Please refer to the documentation to get more information on the supported types."
        )

        with pytest.raises(FeatureStoreException) as exception:
            HopsworksUdf._validate_and_convert_output_types([int, "pd.DatetimeTZDtype"])

        assert (
            str(exception.value)
            == "Output type pd.DatetimeTZDtype is not supported. Please refer to the documentation to get more information on the supported types."
        )

    def test_get_module_imports(self):
        assert HopsworksUdf._get_module_imports(
            "python/tests/test_helpers/transformation_test_helper.py"
        ) == [
            "import pandas as pd",
            "from hsfs.transformation_statistics import TransformationStatistics",
        ]

    def test_extract_source_code(self):
        from .test_helpers.transformation_test_helper import test_function

        assert """import pandas as pd
from hsfs.transformation_statistics import TransformationStatistics
def test_function():
    return True""" == HopsworksUdf._extract_source_code(test_function).strip()

    def test_extract_function_arguments_no_arguments(self):
        from .test_helpers.transformation_test_helper import test_function

        with pytest.raises(FeatureStoreException) as exception:
            HopsworksUdf._extract_function_arguments(test_function)

        assert (
            str(exception.value)
            == "No arguments present in the provided user defined function. Please provide at least one argument in the defined user defined function."
        )

    def test_extract_function_arguments_one_argument(self):
        from .test_helpers.transformation_test_helper import test_function_one_argument

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_one_argument
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name=None)
        ]

    def test_extract_function_arguments_one_argument_with_statistics(self):
        from .test_helpers.transformation_test_helper import (
            test_function_one_argument_with_statistics,
        )

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_one_argument_with_statistics
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name="arg1")
        ]

    def test_extract_function_arguments_one_argument_with_typehint(self):
        from .test_helpers.transformation_test_helper import (
            test_function_one_argument_with_typehints,
        )

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_one_argument_with_typehints
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name=None)
        ]

    def test_extract_function_arguments_one_argument_with_statistics_and_typehints(
        self,
    ):
        from .test_helpers.transformation_test_helper import (
            test_function_one_argument_with_statistics_and_typehints,
        )

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_one_argument_with_statistics_and_typehints
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name="arg1")
        ]

    def test_extract_function_arguments_multiple_argument(self):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument,
        )

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_multiple_argument
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name=None),
            TransformationFeature(feature_name="arg2", statistic_argument_name=None),
        ]

    def test_extract_function_arguments_multiple_argument_with_statistics(self):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_with_statistics,
        )

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_multiple_argument_with_statistics
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name="arg1"),
            TransformationFeature(feature_name="arg2", statistic_argument_name=None),
            TransformationFeature(feature_name="arg3", statistic_argument_name="arg3"),
        ]

    def test_extract_function_arguments_multiple_argument_with_typehints(self):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_with_typehints,
        )

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_multiple_argument_with_typehints
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name=None),
            TransformationFeature(feature_name="arg2", statistic_argument_name=None),
        ]

    def test_extract_function_arguments_multiple_argument_with_statistics_and_typehints(
        self,
    ):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_with_statistics_and_typehints,
        )

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_multiple_argument_with_statistics_and_typehints
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name="arg1"),
            TransformationFeature(feature_name="arg2", statistic_argument_name="arg2"),
        ]

    def test_extract_function_arguments_multiple_argument_with_mixed_statistics_and_typehints(
        self,
    ):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_with_mixed_statistics_and_typehints,
        )

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_multiple_argument_with_mixed_statistics_and_typehints
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name="arg1"),
            TransformationFeature(feature_name="arg2", statistic_argument_name=None),
            TransformationFeature(feature_name="arg3", statistic_argument_name="arg3"),
        ]

    def test_extract_function_arguments_multiple_argument_all_parameter_with_spaces(
        self,
    ):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_all_parameter_with_spaces,
        )

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_multiple_argument_all_parameter_with_spaces
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name="arg1"),
            TransformationFeature(feature_name="arg2", statistic_argument_name="arg2"),
        ]

    def test_extract_function_arguments_multiple_argument_all_parameter_multiline(self):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_all_parameter_multiline,
        )

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_multiple_argument_all_parameter_multiline
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name="arg1"),
            TransformationFeature(feature_name="arg2", statistic_argument_name=None),
            TransformationFeature(feature_name="arg3", statistic_argument_name="arg3"),
        ]

    def test_extract_function_arguments_multiple_argumen_all_parameter_multiline_with_comments(
        self,
    ):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_all_parameter_multiline_with_comments,
        )

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_multiple_argument_all_parameter_multiline_with_comments
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name="arg1"),
            TransformationFeature(feature_name="arg2", statistic_argument_name=None),
            TransformationFeature(feature_name="arg3", statistic_argument_name="arg3"),
        ]

    def test_extract_function_arguments_context_variable(
        self,
    ):
        from .test_helpers.transformation_test_helper import (
            test_function_context_variables,
        )

        function_argument = HopsworksUdf._extract_function_arguments(
            test_function_context_variables
        )

        assert function_argument == [
            TransformationFeature(feature_name="arg1", statistic_argument_name=None),
        ]

    def test_extract_function_arguments_statistics_invalid(self):
        from .test_helpers.transformation_test_helper import (
            test_function_statistics_invalid,
        )

        with pytest.raises(FeatureStoreException) as exception:
            HopsworksUdf._extract_function_arguments(test_function_statistics_invalid)

        assert (
            str(exception.value)
            == "No argument corresponding to statistics parameter 'arg3' present in function definition."
        )

    def test_format_source_code_one_argument(self):
        from .test_helpers.transformation_test_helper import (
            test_function_one_argument,
        )

        function_source = HopsworksUdf._extract_source_code(test_function_one_argument)

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_one_argument(arg1):
\t    pass"""
        )

    def test_format_source_code_one_argument_with_statistics(self):
        from .test_helpers.transformation_test_helper import (
            test_function_one_argument_with_statistics,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_one_argument_with_statistics
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_one_argument_with_statistics(arg1):
\t    pass"""
        )

    def test_format_source_code_one_argument_with_typehints(self):
        from .test_helpers.transformation_test_helper import (
            test_function_one_argument_with_typehints,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_one_argument_with_typehints
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_one_argument_with_typehints(arg1):
\t    pass"""
        )

    def test_format_source_code_one_argument_with_statistics_and_typehints(self):
        from .test_helpers.transformation_test_helper import (
            test_function_one_argument_with_statistics_and_typehints,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_one_argument_with_statistics_and_typehints
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_one_argument_with_statistics_and_typehints(arg1):
\t    pass"""
        )

    def test_format_source_code_multiple_argument(self):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_multiple_argument
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_multiple_argument(arg1, arg2):
\t    pass"""
        )

    def test_format_source_code_multiple_argument_with_statistics(self):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_with_statistics,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_multiple_argument_with_statistics
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_multiple_argument_with_statistics(arg1, arg2, arg3):
\t    pass"""
        )

    def test_format_source_code_multiple_argument_with_typehints(self):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_with_typehints,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_multiple_argument_with_typehints
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_multiple_argument_with_typehints(arg1, arg2):
\t    pass"""
        )

    def test_format_source_code_multiple_argument_with_statistics_and_typehints(self):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_with_statistics_and_typehints,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_multiple_argument_with_statistics_and_typehints
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_multiple_argument_with_statistics_and_typehints(arg1, arg2):
\t    pass"""
        )

    def test_format_source_code_multiple_argument_with_mixed_statistics_and_typehints(
        self,
    ):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_with_mixed_statistics_and_typehints,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_multiple_argument_with_mixed_statistics_and_typehints
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_multiple_argument_with_mixed_statistics_and_typehints(arg1, arg2, arg3):
\t    pass"""
        )

    def test_format_source_code_multiple_argument_all_parameter_with_spaces(self):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_all_parameter_with_spaces,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_multiple_argument_all_parameter_with_spaces
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_multiple_argument_all_parameter_with_spaces(arg1, arg2):
\t    pass"""
        )

    def test_format_source_code_multiple_argument_all_parameter_multiline(self):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_all_parameter_multiline,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_multiple_argument_all_parameter_multiline
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_multiple_argument_all_parameter_multiline(arg1, arg2, arg3):
\t    pass"""
        )

    def test_format_source_code_multiline_with_comments(self):
        from .test_helpers.transformation_test_helper import (
            test_function_multiple_argument_all_parameter_multiline_with_comments,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_multiple_argument_all_parameter_multiline_with_comments
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_multiple_argument_all_parameter_multiline_with_comments(arg1, arg2, arg3):
\t    pass"""
        )

    def test_format_source_code_transformation_statistics_as_default_one_line(self):
        from .test_helpers.transformation_test_helper import (
            test_function_transformation_statistics_as_default_one_line,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_transformation_statistics_as_default_one_line
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_transformation_statistics_as_default_one_line(arg1, arg2):
\t    pass"""
        )

    def test_format_source_code_transformation_statistics_as_default_one_line_return_type(
        self,
    ):
        from .test_helpers.transformation_test_helper import (
            test_function_transformation_statistics_as_default_one_line_return_type,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_transformation_statistics_as_default_one_line_return_type
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_transformation_statistics_as_default_one_line_return_type(arg1, arg2):
\t    pass"""
        )

    def test_format_source_code_transformation_statistics_as_default_multiple_line(
        self,
    ):
        from .test_helpers.transformation_test_helper import (
            test_function_transformation_statistics_as_default_multiple_line,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_transformation_statistics_as_default_multiple_line
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_transformation_statistics_as_default_multiple_line(arg1, arg2):
\t    pass"""
        )

    def test_format_source_code_transformation_statistics_as_default_multiple_line_return_type_spaces(
        self,
    ):
        from .test_helpers.transformation_test_helper import (
            test_function_transformation_statistics_as_default_multiple_line_return_type_spaces,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_transformation_statistics_as_default_multiple_line_return_type_spaces
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_transformation_statistics_as_default_multiple_line_return_type_spaces(arg1, arg2):
\t    pass"""
        )

    def test_format_source_code_transformation_statistics_as_default_multiple_line_return_type_no_spaces(
        self,
    ):
        from .test_helpers.transformation_test_helper import (
            test_function_transformation_statistics_as_default_multiple_line_return_type_no_spaces,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_transformation_statistics_as_default_multiple_line_return_type_no_spaces
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_transformation_statistics_as_default_multiple_line_return_type_no_spaces(arg1, arg2):
\t    pass"""
        )

    def test_format_source_code_context_variable(
        self,
    ):
        from .test_helpers.transformation_test_helper import (
            test_function_context_variables,
        )

        function_source = HopsworksUdf._extract_source_code(
            test_function_context_variables
        )

        formated_source, module_imports = HopsworksUdf._format_source_code(
            function_source
        )

        assert (
            formated_source.strip()
            == """def test_function_context_variables(arg1):
\t    pass"""
        )

    def test_drop_features_one_element(self):
        @udf([int, float, int], drop="col1")
        def test_func(col1, col2, col3):
            return pd.DataFrame(
                {"col1": [col1 + 1], "col2": [col2 + 1], "col3": [col3 + 1]}
            )

        assert test_func.dropped_features == ["col1"]

    def test_drop_features_one_element_prefix(self):
        @udf([int, float, int], drop="col1")
        def test_func(col1, col2, col3):
            return pd.DataFrame(
                {"col1": [col1 + 1], "col2": [col2 + 1], "col3": [col3 + 1]}
            )

        test_func._feature_name_prefix = "prefix_"

        assert test_func._dropped_features == ["col1"]
        assert test_func.dropped_features == ["prefix_col1"]

    def test_drop_features_multiple_element(self):
        @udf([int, float, int], drop=["col1", "col2"])
        def test_func(col1, col2, col3):
            return pd.DataFrame(
                {"col1": [col1 + 1], "col2": [col2 + 1], "col3": [col3 + 1]}
            )

        assert test_func.dropped_features == ["col1", "col2"]

    def test_drop_features_multiple_element_prefix(self):
        @udf([int, float, int], drop=["col1", "col2"])
        def test_func(col1, col2, col3):
            return pd.DataFrame(
                {"col1": [col1 + 1], "col2": [col2 + 1], "col3": [col3 + 1]}
            )

        test_func._feature_name_prefix = "prefix_"

        assert test_func._dropped_features == ["col1", "col2"]
        assert test_func.dropped_features == ["prefix_col1", "prefix_col2"]

    def test_drop_features_invalid(self):
        with pytest.raises(FeatureStoreException) as exp:

            @udf([int, float, int], drop=["col1", "invalid_col"])
            def test_func(col1, col2, col3):
                return pd.DataFrame(
                    {"col1": [col1 + 1], "col2": [col2 + 1], "col3": [col3 + 1]}
                )

        assert (
            str(exp.value)
            == "Cannot drop features 'invalid_col' as they are not features given as arguments in the defined UDF."
        )

    def test_create_pandas_udf_return_schema_from_list_one_output_type(self):
        @udf(int)
        def test_func(col1):
            return col1 + 1

        assert test_func._create_pandas_udf_return_schema_from_list() == "bigint"

    def test_create_pandas_udf_return_schema_from_list_one_argument_multiple_output_type(
        self,
    ):
        @udf([int, float, str, date, datetime, time, bool])
        def test_func(col1):
            return pd.DataFrame(
                {
                    "col1": [col1 + 1],
                    "col2": [col1 + 1],
                    "col3": [col1 + 1],
                    "col4": [col1 + 1],
                    "col5": [col1 + 1],
                    "col6": [True],
                }
            )

        test_func.output_column_names = [
            "test_func_col1_0",
            "test_func_col1_1",
            "test_func_col1_2",
            "test_func_col1_3",
            "test_func_col1_4",
            "test_func_col1_5",
            "test_func_col1_6",
        ]

        assert (
            test_func._create_pandas_udf_return_schema_from_list()
            == "`test_func_col1_0` bigint, `test_func_col1_1` double, `test_func_col1_2` string, `test_func_col1_3` date, `test_func_col1_4` timestamp, `test_func_col1_5` timestamp, `test_func_col1_6` boolean"
        )

    def test_pandas_udf_wrapper_single_output(self):
        test_dataframe = pd.DataFrame({"col1": [1, 2, 3, 4]})

        @udf(int)
        def test_func(col1):
            return col1 + 1

        test_func.output_column_names = ["test_func_col1_"]
        renaming_wrapper_function = test_func.pandas_udf_wrapper()

        result = renaming_wrapper_function(test_dataframe["col1"])

        assert result.name == "test_func_col1_"
        assert result.values.tolist() == [2, 3, 4, 5]

    def test_pandas_udf_wrapper_context_variables(self):
        test_dataframe = pd.DataFrame({"col1": [1, 2, 3, 4]})

        @udf(int)
        def test_func(col1, context):
            return col1 + context["test_value"]

        test_func.output_column_names = ["test_func_col1_"]
        test_func.transformation_context = {"test_value": 200}
        renaming_wrapper_function = test_func.pandas_udf_wrapper()

        result = renaming_wrapper_function(test_dataframe["col1"])

        assert result.name == "test_func_col1_"
        assert result.values.tolist() == [201, 202, 203, 204]

    def test_python_udf_wrapper_single_output(self):
        test_dataframe = pd.DataFrame({"col1": [1, 2, 3, 4]})

        @udf(int)
        def test_func(col1):
            return col1 + 1

        test_func.output_column_names = ["test_func_col1_"]
        wrapper_function = test_func.python_udf_wrapper(rename_outputs=False)

        result = test_dataframe.apply(
            lambda x: wrapper_function(x["col1"]), axis=1, result_type="expand"
        )

        assert result.values.tolist() == [2, 3, 4, 5]

    def test_python_udf_wrapper_context_variables(self):
        test_dataframe = pd.DataFrame({"col1": [1, 2, 3, 4]})

        @udf(int)
        def test_func(col1, context):
            return col1 + context["test_value"]

        test_func.transformation_context = {"test_value": 100}
        test_func.output_column_names = ["test_func_col1_"]
        wrapper_function = test_func.python_udf_wrapper(rename_outputs=False)

        result = test_dataframe.apply(
            lambda x: wrapper_function(x["col1"]), axis=1, result_type="expand"
        )

        assert result.values.tolist() == [101, 102, 103, 104]

    def test_pandas_udf_wrapper_multiple_output(self):
        @udf([int, float])
        def test_func(col1, col2):
            return pd.DataFrame({"out1": col1 + 1, "out2": col2 + 2})

        test_func.output_column_names = [
            "test_func_col1_col2_0",
            "test_func_col1_col2_1",
        ]
        renaming_wrapper_function = test_func.pandas_udf_wrapper()

        test_dataframe = pd.DataFrame(
            {"column1": [1, 2, 3, 4], "column2": [10, 20, 30, 40]}
        )

        result = renaming_wrapper_function(
            test_dataframe["column1"], test_dataframe["column2"]
        )

        assert all(result.columns == ["test_func_col1_col2_0", "test_func_col1_col2_1"])
        assert result.values.tolist() == [[2, 12], [3, 22], [4, 32], [5, 42]]

    def test_python_udf_wrapper_multiple_output(self):
        @udf([int, float], mode="python")
        def test_func(col1, col2):
            return col1 + 1, col2 + 2

        test_func.output_column_names = [
            "test_func_col1_col2_0",
            "test_func_col1_col2_1",
        ]
        wrapper_function = test_func.python_udf_wrapper(rename_outputs=False)

        test_dataframe = pd.DataFrame(
            {"column1": [1, 2, 3, 4], "column2": [10, 20, 30, 40]}
        )

        result = test_dataframe.apply(
            lambda x: wrapper_function(x["column1"], x["column2"]),
            axis=1,
            result_type="expand",
        )

        assert result.values.tolist() == [[2, 12], [3, 22], [4, 32], [5, 42]]

        renaming_wrapper_function = test_func.python_udf_wrapper(rename_outputs=True)

        result = test_dataframe.apply(
            lambda x: renaming_wrapper_function(x["column1"], x["column2"]),
            axis=1,
            result_type="expand",
        )

        assert all(result.columns == ["test_func_col1_col2_0", "test_func_col1_col2_1"])
        assert result.values.tolist() == [[2, 12], [3, 22], [4, 32], [5, 42]]

    def test_get_udf_spark_engine_default_mode_training(self, mocker):
        mocker.patch("hsfs.engine.get_type", return_value="spark")

        @udf(return_type=int)
        def test(feature):
            pass

        pandas_udf_mocker = mocker.patch("pyspark.sql.functions.pandas_udf")
        python_udf_mocker = mocker.patch("pyspark.sql.functions.udf")
        pandas_wrapper_mocker = mocker.patch.object(test, "pandas_udf_wrapper")
        python_wrapper_mocker = mocker.patch.object(test, "python_udf_wrapper")

        _ = test.get_udf(online=False)

        assert pandas_udf_mocker.call_count == 1
        assert python_udf_mocker.call_count == 0
        assert pandas_wrapper_mocker.call_count == 1
        assert python_wrapper_mocker.call_count == 0

    def test_get_udf_spark_engine_default_mode_inference(self, mocker):
        mocker.patch("hsfs.engine.get_type", return_value="spark")

        @udf(return_type=int)
        def test(feature):
            pass

        pandas_udf_mocker = mocker.patch("pyspark.sql.functions.pandas_udf")
        python_udf_mocker = mocker.patch("pyspark.sql.functions.udf")
        pandas_wrapper_mocker = mocker.patch.object(test, "pandas_udf_wrapper")
        python_wrapper_mocker = mocker.patch.object(test, "python_udf_wrapper")

        _ = test.get_udf(online=True)

        assert pandas_udf_mocker.call_count == 0
        assert python_udf_mocker.call_count == 0
        assert pandas_wrapper_mocker.call_count == 0
        assert python_wrapper_mocker.call_count == 1

    def test_get_udf_spark_engine_python_mode_training(self, mocker):
        mocker.patch("hsfs.engine.get_type", return_value="spark")

        @udf(return_type=int, mode="python")
        def test(feature):
            pass

        pandas_udf_mocker = mocker.patch("pyspark.sql.functions.pandas_udf")
        python_udf_mocker = mocker.patch("pyspark.sql.functions.udf")
        pandas_wrapper_mocker = mocker.patch.object(test, "pandas_udf_wrapper")
        python_wrapper_mocker = mocker.patch.object(test, "python_udf_wrapper")

        _ = test.get_udf(online=False)

        assert pandas_udf_mocker.call_count == 0
        assert python_udf_mocker.call_count == 1
        assert pandas_wrapper_mocker.call_count == 0
        assert python_wrapper_mocker.call_count == 1

    def test_get_udf_spark_engine_python_mode_inference(self, mocker):
        mocker.patch("hsfs.engine.get_type", return_value="spark")

        @udf(return_type=int, mode="python")
        def test(feature):
            pass

        pandas_udf_mocker = mocker.patch("pyspark.sql.functions.pandas_udf")
        python_udf_mocker = mocker.patch("pyspark.sql.functions.udf")
        pandas_wrapper_mocker = mocker.patch.object(test, "pandas_udf_wrapper")
        python_wrapper_mocker = mocker.patch.object(test, "python_udf_wrapper")

        _ = test.get_udf(online=True)

        assert pandas_udf_mocker.call_count == 0
        assert python_udf_mocker.call_count == 0
        assert pandas_wrapper_mocker.call_count == 0
        assert python_wrapper_mocker.call_count == 1

    def test_get_udf_spark_engine_pandas_mode_training(self, mocker):
        mocker.patch("hsfs.engine.get_type", return_value="spark")

        @udf(return_type=int, mode="pandas")
        def test(feature):
            pass

        pandas_udf_mocker = mocker.patch("pyspark.sql.functions.pandas_udf")
        python_udf_mocker = mocker.patch("pyspark.sql.functions.udf")
        pandas_wrapper_mocker = mocker.patch.object(test, "pandas_udf_wrapper")
        python_wrapper_mocker = mocker.patch.object(test, "python_udf_wrapper")

        _ = test.get_udf(online=False)

        assert pandas_udf_mocker.call_count == 1
        assert python_udf_mocker.call_count == 0
        assert pandas_wrapper_mocker.call_count == 1
        assert python_wrapper_mocker.call_count == 0

    def test_get_udf_spark_engine_pandas_mode_inference(self, mocker):
        mocker.patch("hsfs.engine.get_type", return_value="spark")

        @udf(return_type=int, mode="pandas")
        def test(feature):
            pass

        pandas_udf_mocker = mocker.patch("pyspark.sql.functions.pandas_udf")
        python_udf_mocker = mocker.patch("pyspark.sql.functions.udf")
        pandas_wrapper_mocker = mocker.patch.object(test, "pandas_udf_wrapper")
        python_wrapper_mocker = mocker.patch.object(test, "python_udf_wrapper")

        _ = test.get_udf(online=True)

        assert pandas_udf_mocker.call_count == 0
        assert python_udf_mocker.call_count == 0
        assert pandas_wrapper_mocker.call_count == 1
        assert python_wrapper_mocker.call_count == 0

    def test_get_udf_python_engine_default_mode_training(self, mocker):
        mocker.patch("hsfs.engine.get_type", return_value="python")

        @udf(return_type=int)
        def test(feature):
            pass

        pandas_udf_mocker = mocker.patch("pyspark.sql.functions.pandas_udf")
        python_udf_mocker = mocker.patch("pyspark.sql.functions.udf")
        pandas_wrapper_mocker = mocker.patch.object(test, "pandas_udf_wrapper")
        python_wrapper_mocker = mocker.patch.object(test, "python_udf_wrapper")

        _ = test.get_udf(online=False)

        assert pandas_udf_mocker.call_count == 0
        assert python_udf_mocker.call_count == 0
        assert pandas_wrapper_mocker.call_count == 1
        assert python_wrapper_mocker.call_count == 0

    def test_get_udf_python_engine_default_mode_inference(self, mocker):
        mocker.patch("hsfs.engine.get_type", return_value="python")

        @udf(return_type=int)
        def test(feature):
            pass

        pandas_udf_mocker = mocker.patch("pyspark.sql.functions.pandas_udf")
        python_udf_mocker = mocker.patch("pyspark.sql.functions.udf")
        pandas_wrapper_mocker = mocker.patch.object(test, "pandas_udf_wrapper")
        python_wrapper_mocker = mocker.patch.object(test, "python_udf_wrapper")

        _ = test.get_udf(online=True)

        assert pandas_udf_mocker.call_count == 0
        assert python_udf_mocker.call_count == 0
        assert pandas_wrapper_mocker.call_count == 0
        assert python_wrapper_mocker.call_count == 1

    def test_get_udf_python_engine_python_mode_training(self, mocker):
        mocker.patch("hsfs.engine.get_type", return_value="python")

        @udf(return_type=int, mode="python")
        def test(feature):
            pass

        pandas_udf_mocker = mocker.patch("pyspark.sql.functions.pandas_udf")
        python_udf_mocker = mocker.patch("pyspark.sql.functions.udf")
        pandas_wrapper_mocker = mocker.patch.object(test, "pandas_udf_wrapper")
        python_wrapper_mocker = mocker.patch.object(test, "python_udf_wrapper")

        _ = test.get_udf(online=False)

        assert pandas_udf_mocker.call_count == 0
        assert python_udf_mocker.call_count == 0
        assert pandas_wrapper_mocker.call_count == 0
        assert python_wrapper_mocker.call_count == 1

    def test_get_udf_python_engine_python_mode_inference(self, mocker):
        mocker.patch("hsfs.engine.get_type", return_value="python")

        @udf(return_type=int, mode="python")
        def test(feature):
            pass

        pandas_udf_mocker = mocker.patch("pyspark.sql.functions.pandas_udf")
        python_udf_mocker = mocker.patch("pyspark.sql.functions.udf")
        pandas_wrapper_mocker = mocker.patch.object(test, "pandas_udf_wrapper")
        python_wrapper_mocker = mocker.patch.object(test, "python_udf_wrapper")

        _ = test.get_udf(online=True)

        assert pandas_udf_mocker.call_count == 0
        assert python_udf_mocker.call_count == 0
        assert pandas_wrapper_mocker.call_count == 0
        assert python_wrapper_mocker.call_count == 1

    def test_get_udf_python_engine_pandas_mode_training(self, mocker):
        mocker.patch("hsfs.engine.get_type", return_value="python")

        @udf(return_type=int, mode="pandas")
        def test(feature):
            pass

        pandas_udf_mocker = mocker.patch("pyspark.sql.functions.pandas_udf")
        python_udf_mocker = mocker.patch("pyspark.sql.functions.udf")
        pandas_wrapper_mocker = mocker.patch.object(test, "pandas_udf_wrapper")
        python_wrapper_mocker = mocker.patch.object(test, "python_udf_wrapper")

        _ = test.get_udf(online=False)

        assert pandas_udf_mocker.call_count == 0
        assert python_udf_mocker.call_count == 0
        assert pandas_wrapper_mocker.call_count == 1
        assert python_wrapper_mocker.call_count == 0

    def test_get_udf_python_engine_pandas_mode_inference(self, mocker):
        mocker.patch("hsfs.engine.get_type", return_value="python")

        @udf(return_type=int, mode="pandas")
        def test(feature):
            pass

        pandas_udf_mocker = mocker.patch("pyspark.sql.functions.pandas_udf")
        python_udf_mocker = mocker.patch("pyspark.sql.functions.udf")
        pandas_wrapper_mocker = mocker.patch.object(test, "pandas_udf_wrapper")
        python_wrapper_mocker = mocker.patch.object(test, "python_udf_wrapper")

        _ = test.get_udf(online=True)

        assert pandas_udf_mocker.call_count == 0
        assert python_udf_mocker.call_count == 0
        assert pandas_wrapper_mocker.call_count == 1
        assert python_wrapper_mocker.call_count == 0

    def test_HopsworkUDf_call_one_argument(self):
        @udf(int)
        def test_func(col1):
            return col1 + 1

        assert test_func.transformation_features == ["col1"]
        assert test_func.statistics_features == []

        assert test_func("new_feature").transformation_features == ["new_feature"]
        assert test_func("new_feature").statistics_features == []

        # Test with prefix
        test_func._feature_name_prefix = "prefix_"
        assert test_func.transformation_features == ["prefix_col1"]
        assert test_func.statistics_features == []

        assert test_func("new_feature").transformation_features == [
            "prefix_new_feature"
        ]
        assert test_func("new_feature").statistics_features == []

    def test_HopsworkUDf_call_one_argument_statistics(self):
        from hsfs.transformation_statistics import TransformationStatistics

        stats = TransformationStatistics("col1")

        @udf(int)
        def test_func(col1, statistics=stats):
            return col1 + statistics.col1.mean

        assert test_func.transformation_features == ["col1"]
        assert test_func.statistics_features == ["col1"]
        assert test_func._statistics_argument_names == ["col1"]

        assert test_func("new_feature").transformation_features == ["new_feature"]
        assert test_func("new_feature").statistics_features == ["new_feature"]
        assert test_func("new_feature")._statistics_argument_names == ["col1"]

        # Test with prefix
        test_func._feature_name_prefix = "prefix_"
        assert test_func.transformation_features == ["prefix_col1"]
        assert test_func.statistics_features == ["col1"]
        assert test_func._statistics_argument_names == ["col1"]

        assert test_func("new_feature").transformation_features == [
            "prefix_new_feature"
        ]
        assert test_func("new_feature").statistics_features == ["new_feature"]
        assert test_func("new_feature")._statistics_argument_names == ["col1"]

    def test_HopsworkUDf_call_multiple_argument_statistics(self):
        from hsfs.transformation_statistics import TransformationStatistics

        stats = TransformationStatistics("col1", "col3")

        @udf(int)
        def test_func(col1, col2, col3, statistics=stats):
            return col1 + statistics.col1.mean + statistics.col3.mean

        assert test_func.transformation_features == ["col1", "col2", "col3"]
        assert test_func.statistics_features == ["col1", "col3"]

        assert test_func("f1", "f2", "f3").transformation_features == ["f1", "f2", "f3"]
        assert test_func("f1", "f2", "f3").statistics_features == ["f1", "f3"]
        assert test_func("f1", "f2", "f3")._statistics_argument_names == [
            "col1",
            "col3",
        ]

    def test_validate_and_convert_drop_features(self):
        dropped_features = "feature1"
        transformation_feature = ["feature1", "feature2"]
        feature_name_prefix = None

        dropped_features = HopsworksUdf._validate_and_convert_drop_features(
            dropped_features, transformation_feature, feature_name_prefix
        )

        assert dropped_features == ["feature1"]

    def test_validate_and_convert_drop_features_dropped_list(self):
        dropped_features = ["feature1", "feature2"]
        transformation_feature = ["feature1", "feature2", "feature3"]
        feature_name_prefix = None

        dropped_features = HopsworksUdf._validate_and_convert_drop_features(
            dropped_features, transformation_feature, feature_name_prefix
        )

        assert dropped_features == ["feature1", "feature2"]

    def test_validate_and_convert_drop_features_dropped_invalid(self):
        dropped_features = "feature4"
        transformation_feature = ["feature1", "feature2", "feature3"]
        feature_name_prefix = None

        with pytest.raises(FeatureStoreException) as exp:
            HopsworksUdf._validate_and_convert_drop_features(
                dropped_features, transformation_feature, feature_name_prefix
            )

        assert (
            str(exp.value)
            == "Cannot drop features 'feature4' as they are not features given as arguments in the defined UDF."
        )

    def test_validate_and_convert_drop_features_dropped_invalid_list(self):
        dropped_features = ["feature4", "feature5"]
        transformation_feature = ["feature1", "feature2", "feature3"]
        feature_name_prefix = None

        with pytest.raises(FeatureStoreException) as exp:
            HopsworksUdf._validate_and_convert_drop_features(
                dropped_features, transformation_feature, feature_name_prefix
            )

        assert (
            str(exp.value)
            == "Cannot drop features 'feature4', 'feature5' as they are not features given as arguments in the defined UDF."
        )

    def test_validate_and_convert_drop_features_dropped_list_prefix(self):
        dropped_features = ["feature1", "feature2"]
        transformation_feature = ["test_feature1", "test_feature2", "test_feature3"]
        feature_name_prefix = "test_"

        dropped_features = HopsworksUdf._validate_and_convert_drop_features(
            dropped_features, transformation_feature, feature_name_prefix
        )

        assert dropped_features == ["feature1", "feature2"]

    def test_validate_and_convert_drop_features_dropped_prefix_invalid(self):
        dropped_features = ["feature1", "feature2"]
        transformation_feature = ["feature1", "feature2", "feature3"]
        feature_name_prefix = "test_"

        with pytest.raises(FeatureStoreException) as exp:
            HopsworksUdf._validate_and_convert_drop_features(
                dropped_features, transformation_feature, feature_name_prefix
            )

        assert (
            str(exp.value)
            == "Cannot drop features 'test_feature1', 'test_feature2' as they are not features given as arguments in the defined UDF."
        )

    def test_alias_one_output(self):
        @udf(int)
        def add_one(feature):
            return feature + 1

        add_one = add_one.alias("feature_plus_one")

        assert add_one.output_column_names == ["feature_plus_one"]

    def test_alias_one_output_list(self):
        @udf(int)
        def add_one(feature):
            return feature + 1

        add_one = add_one.alias(["feature_plus_one"])

        assert add_one.output_column_names == ["feature_plus_one"]

    def test_alias_multiple_output(self):
        @udf([int, int])
        def add_and_sub(feature):
            return feature + 1, feature - 1

        add_one = add_and_sub.alias("feature_plus_one", "feature_minus_one")

        assert add_one.output_column_names == ["feature_plus_one", "feature_minus_one"]

    def test_alias_multiple_output_list(self):
        @udf([int, int])
        def add_and_sub(feature):
            return feature + 1, feature - 1

        add_one = add_and_sub.alias(["feature_plus_one", "feature_minus_one"])

        assert add_one.output_column_names == ["feature_plus_one", "feature_minus_one"]

    def test_alias_invalid_number_column_names(self):
        @udf([int, int])
        def add_and_sub(feature):
            return feature + 1, feature - 1

        with pytest.raises(FeatureStoreException) as exp:
            add_and_sub.alias(["feature_plus_one", "feature_minus_one", "invalid_col"])

        assert (
            str(exp.value)
            == "The number of output feature names provided does not match the number of features returned by the transformation function 'add_and_sub(feature)'. Pease provide exactly 2 feature name(s) to match the output."
        )

    def test_alias_invalid_type(self):
        @udf([int, int])
        def add_and_sub(feature):
            return feature + 1, feature - 1

        with pytest.raises(FeatureStoreException) as exp:
            add_and_sub.alias("feature_plus_one", {"name": "col1"})

        assert (
            str(exp.value)
            == "Invalid output feature names provided for the transformation function 'add_and_sub(feature)'. Please ensure all arguments are strings."
        )

    def test_alias_duplicates(self):
        @udf([int, int])
        def add_and_sub(feature):
            return feature + 1, feature - 1

        with pytest.raises(FeatureStoreException) as exp:
            add_and_sub.alias("feature_plus_one", "feature_plus_one")

        assert (
            str(exp.value)
            == "Duplicate output feature names provided for the transformation function 'add_and_sub(feature)'. Please ensure all arguments names are unique."
        )

    def test_call_and_alias(self):
        @udf([int, int])
        def add_and_sub(feature):
            return feature + 1, feature - 1

        add_one = add_and_sub("feature2").alias(
            ["feature_plus_one", "feature_minus_one"]
        )

        assert add_one.output_column_names == ["feature_plus_one", "feature_minus_one"]
        assert add_one.transformation_features == ["feature2"]

    def test_alias_invalid_length(self):
        @udf([int, int])
        def add_and_sub(feature):
            return feature + 1, feature - 1

        with pytest.raises(FeatureStoreException) as exp:
            add_and_sub.alias(["invalid" * 10, "feature_minus_one"])

        assert (
            str(exp.value)
            == "Invalid output feature names specified for the transformation function 'add_and_sub(feature)'. Please provide names shorter than 63 characters."
        )

    def test_invalid_transformation_context(self):
        @udf(int)
        def test_func(feature, context):
            return feature + context["test_value"]

        with pytest.raises(FeatureStoreException) as exp:
            test_func.transformation_context = "invalid_context"

        exp.match("Transformation context variable must be passed as dictionary.")

    def test_prepare_transformation_function_scope_no_kwargs_no_statistics_no_context(
        self,
    ):
        @udf(int)
        def test_func(feature):
            return feature + 1

        # Setting the output column names to test the scope as they would always be set before scope is prepared.
        test_func._output_column_names = ["test_func_feature_"]

        scope = test_func._prepare_transformation_function_scope()

        assert scope["_output_col_names"] == ["test_func_feature_"]
        assert "_output_col_names" in scope.keys()

    def test_prepare_transformation_function_scope_no_kwargs_no_statistics_context(
        self,
    ):
        @udf(int)
        def test_func(feature, context):
            return feature + 1

        # Setting the output column names to test the scope, transformation context as they would always be set before scope is prepared.
        test_func._output_column_names = ["test_func_feature_"]
        test_func.transformation_context = {"test_value": 100}

        scope = test_func._prepare_transformation_function_scope()

        print(scope)

        assert scope["_output_col_names"] == ["test_func_feature_"]
        assert scope["context"] == {"test_value": 100}
        assert all(
            value in scope.keys()
            for value in {
                "_output_col_names",
                "context",
            }
        )

    def test_prepare_transformation_function_scope_no_kwargs_statistics_context(self):
        @udf(int)
        def test_func(feature, context):
            return feature + 1

        # Setting the output column names to test the scope, transformation context and statistics as they would always be set before scope is prepared.
        test_func._output_column_names = ["test_func_feature_"]
        test_func.transformation_context = {"test_value": 100}
        test_func._statistics = 10

        scope = test_func._prepare_transformation_function_scope()

        assert scope["_output_col_names"] == ["test_func_feature_"]
        assert scope["context"] == {"test_value": 100}
        assert scope["statistics"] == 10
        assert all(
            value in scope.keys()
            for value in {"_output_col_names", "context", "statistics"}
        )

    def test_prepare_transformation_function_scope_kwargs_statistics_context(self):
        @udf(int)
        def test_func(feature, context):
            return feature + 1

        # Setting the output column names to test the scope, transformation context and statistics as they would always be set before scope is prepared.
        test_func._output_column_names = ["test_func_feature_"]
        test_func.transformation_context = {"test_value": 100}
        test_func._statistics = 10

        scope = test_func._prepare_transformation_function_scope(test="values")

        assert scope["_output_col_names"] == ["test_func_feature_"]
        assert scope["context"] == {"test_value": 100}
        assert scope["statistics"] == 10
        assert scope["test"] == "values"
        assert all(
            value in scope.keys()
            for value in {"_output_col_names", "context", "statistics", "test"}
        )



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json

import humps
from hsfs import util


class FeatureGroupCommit:
    NOT_FOUND_ERROR_CODE = 270227

    def __init__(
        self,
        commitid=None,
        commit_date_string=None,
        rows_inserted=None,
        rows_updated=None,
        rows_deleted=None,
        validation_id=None,
        commit_time=None,
        archived=None,
        last_active_commit_time=None,
        items=None,
        count=None,
        href=None,
        **kwargs,
    ):
        self._commitid = commitid
        self._commit_date_string = commit_date_string
        self._commit_time = commit_time
        self._rows_inserted = rows_inserted
        self._rows_updated = rows_updated
        self._rows_deleted = rows_deleted
        self._validation_id = validation_id
        self._archived = archived
        self._last_active_commit_time = last_active_commit_time

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return [cls(**commit_dto) for commit_dto in json_decamelized["items"]]

    def update_from_response_json(self, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        _ = json_decamelized.pop("href")
        self.__init__(**json_decamelized)
        return self

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        return {
            "commitID": self._commitid,
            "commitDateString": self._commit_date_string,
            "commitTime": self._commit_time,
            "rowsInserted": self._rows_inserted,
            "rowsUpdated": self._rows_updated,
            "rowsDeleted": self._rows_deleted,
            "validationId": self._validation_id,
            "archived": self._archived,
            "lastActiveCommitTime": self._last_active_commit_time,
        }

    @property
    def commitid(self):
        return self._commitid

    @property
    def commit_date_string(self):
        return self._commit_date_string

    @property
    def commit_time(self):
        return self._commit_time

    @property
    def rows_inserted(self):
        return self._rows_inserted

    @property
    def rows_updated(self):
        return self._rows_updated

    @property
    def rows_deleted(self):
        return self._rows_deleted

    @property
    def validation_id(self):
        return self._validation_id

    @property
    def archived(self):
        return self._archived

    @property
    def last_active_commit_time(self):
        return self._last_active_commit_time

    @commitid.setter
    def commitid(self, commitid):
        self._commitid = commitid

    @commit_time.setter
    def commit_time(self, commit_time):
        self._commit_time = commit_time

    @rows_inserted.setter
    def rows_inserted(self, rows_inserted):
        self._rows_inserted = rows_inserted

    @rows_updated.setter
    def rows_updated(self, rows_updated):
        self._rows_updated = rows_updated

    @rows_deleted.setter
    def rows_deleted(self, rows_deleted):
        self._rows_deleted = rows_deleted

    @validation_id.setter
    def validation_id(self, validation_id):
        self._validation_id = validation_id

    @last_active_commit_time.setter
    def last_active_commit_time(self, last_active_commit_time):
        self._last_active_commit_time = last_active_commit_time



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
from typing import Optional

import humps
from hopsworks_common import util


class EmailConfig:
    def __init__(
        self,
        to=None,
        send_resolved: bool = False,
    ):
        if not isinstance(send_resolved, bool):
            raise TypeError("send_resolved must be set to a boolean value")
        if not isinstance(to, str):
            raise TypeError("to must be set to a string value")
        self._send_resolved = send_resolved
        self._to = to

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """return the email config as a dictionary"""
        return {
            "to": self._to,
            "sendResolved": self._send_resolved,
        }


class _EmailConfig:
    def __init__(
        self,
        to=None,
        send_resolved: bool = False,
        **kwargs,
    ):
        self._send_resolved = send_resolved
        self._to = to

    @classmethod
    def from_response_json(cls, json_dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    @property
    def to(self) -> Optional[str]:
        """return the email address of the email config"""
        return self._to

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """return the email config as a dictionary"""
        return {
            "to": self._to,
            "sendResolved": self._send_resolved,
        }

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"EmailConfig({self._to!r})"


class SlackConfig:
    def __init__(
        self,
        channel=None,
        send_resolved: bool = False,
    ):
        if not isinstance(send_resolved, bool):
            raise TypeError("send_resolved must be set to a boolean value")
        if not isinstance(channel, str):
            raise TypeError("to must be set to a string value")
        self._send_resolved = send_resolved
        self._channel = channel

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """return the slack config as a dictionary"""
        return {
            "channel": self._channel,
            "sendResolved": self._send_resolved,
        }


class _SlackConfig:
    def __init__(
        self,
        channel=None,
        send_resolved: bool = False,
        **kwargs,
    ):
        self._send_resolved = send_resolved
        self._channel = channel

    @classmethod
    def from_response_json(cls, json_dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    @property
    def channel(self) -> Optional[str]:
        """return the channel of the slack config"""
        return self._channel

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """return the slack config as a dictionary"""
        return {
            "channel": self._channel,
            "sendResolved": self._send_resolved,
        }

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"SlackConfig({self._channel!r})"


class PagerDutyConfig:
    def __init__(
        self,
        service_key=None,
        routing_key=None,
        send_resolved: bool = False,
    ):
        if not isinstance(send_resolved, bool):
            raise TypeError("send_resolved must be set to a boolean value")
        if not isinstance(service_key, str):
            raise TypeError("service_key must be set to a string value")
        if not isinstance(routing_key, str):
            raise TypeError("routing_key must be set to a string value")
        self._send_resolved = send_resolved
        self._service_key = service_key
        self._routing_key = routing_key

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """return the pager duty config as a dictionary"""
        return {
            "serviceKey": self._service_key,
            "routingKey": self._routing_key,
            "sendResolved": self._send_resolved,
        }


class _PagerDutyConfig:
    def __init__(
        self,
        service_key=None,
        routing_key=None,
        send_resolved: bool = False,
        **kwargs,
    ):
        self._send_resolved = send_resolved
        self._service_key = service_key
        self._routing_key = routing_key

    @classmethod
    def from_response_json(cls, json_dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    @property
    def service_key(self) -> Optional[str]:
        """return the service key of the pager duty config"""
        return self._service_key

    @property
    def routing_key(self) -> Optional[str]:
        """return the routing key of the pager duty config"""
        return self._routing_key

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """return the pager duty config as a dictionary"""
        return {
            "serviceKey": self._service_key,
            "routingKey": self._routing_key,
            "sendResolved": self._send_resolved,
        }

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"PagerDutyConfig({self._service_key!r}, {self._routing_key!r})"


class WebhookConfig:
    def __init__(
        self,
        url=None,
        send_resolved: bool = False,
    ):
        if not isinstance(send_resolved, bool):
            raise TypeError("send_resolved must be set to a boolean value")
        if not isinstance(url, str):
            raise TypeError("url must be set to a string value")
        self._send_resolved = send_resolved
        self._url = url

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """return the webhook config as a dictionary"""
        return {
            "url": self._url,
            "sendResolved": self._send_resolved,
        }


class _WebhookConfig:
    def __init__(
        self,
        url=None,
        send_resolved: bool = False,
        **kwargs,
    ):
        self._send_resolved = send_resolved
        self._url = url

    @classmethod
    def from_response_json(cls, json_dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    @property
    def url(self) -> Optional[str]:
        """return the url of the webhook config"""
        return self._url

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """return the webhook config as a dictionary"""
        return {
            "url": self._url,
            "sendResolved": self._send_resolved,
        }

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"WebhookConfig({self._url!r})"


class AlertReceiver:
    NOT_FOUND_ERROR_CODE = 390003
    def __init__(
        self,
        name=None,
        email_configs=None,
        slack_configs=None,
        pager_duty_configs=None,
        webhook_configs=None,
        **kwargs,
    ):
        self._name = name
        self._email_configs = (
            [_EmailConfig.from_response_json(config) for config in email_configs]
            if email_configs
            else None
        )
        self._slack_configs = (
            [_SlackConfig.from_response_json(config) for config in slack_configs]
            if slack_configs
            else None
        )
        self._pager_duty_configs = (
            [
                _PagerDutyConfig.from_response_json(config)
                for config in pager_duty_configs
            ]
            if pager_duty_configs
            else None
        )
        self._webhook_configs = (
            [_WebhookConfig.from_response_json(config) for config in webhook_configs]
            if webhook_configs
            else None
        )

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if "count" in json_decamelized:
            if "items" in json_decamelized:
                return [cls(**receiver) for receiver in json_decamelized["items"]]
            else:
                return []
        else:
            return cls(**json_decamelized)

    @property
    def name(self) -> Optional[str]:
        """return the name of the alert receiver"""
        return self._name

    @property
    def email_configs(self) -> Optional[list]:
        """return the email configs of the alert receiver"""
        return self._email_configs

    @property
    def slack_configs(self) -> Optional[list]:
        """return the slack configs of the alert receiver"""
        return self._slack_configs

    @property
    def pager_duty_configs(self) -> Optional[list]:
        """return the pager duty configs of the alert receiver"""
        return self._pager_duty_configs

    @property
    def webhook_configs(self) -> Optional[list]:
        """return the webhook configs of the alert receiver"""
        return self._webhook_configs

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """return the alert receiver as a dictionary"""
        dict = {
            "name": self._name,
        }
        if self._email_configs:
            dict["email_configs"] = [config.to_dict() for config in self._email_configs]
        if self._slack_configs:
            dict["slack_configs"] = [config.to_dict() for config in self._slack_configs]
        if self._pager_duty_configs:
            dict["pager_duty_configs"] = [
                config.to_dict() for config in self._pager_duty_configs
            ]
        if self._webhook_configs:
            dict["webhook_configs"] = [
                config.to_dict() for config in self._webhook_configs
            ]
        return dict

    def __str__(self):
        return self.json()

    def _get_config(self):
        if self._email_configs:
            return self._email_configs
        if self._slack_configs:
            return self._slack_configs
        if self._pager_duty_configs:
            return self._pager_duty_configs
        if self._webhook_configs:
            return self._webhook_configs

    def __repr__(self):
        return f"AlertReceiver({self._name!r}, {self._get_config()!r})"



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import logging
import time

from hopsworks_common.client.exceptions import GitException
from hopsworks_common.core import git_op_execution_api


class GitEngine:
    def __init__(self):
        self._git_op_execution_api = git_op_execution_api.GitOpExecutionApi()
        self._log = logging.getLogger(__name__)

    def execute_op_blocking(self, git_op, command):
        """Poll a git execution status until it reaches a terminal state
        :param git_op: git execution to monitor
        :type git_op: GitOpExecution
        :param command: git operation running
        :type command: str
        :return: The final GitOpExecution object
        :rtype: GitOpExecution
        """

        while git_op.success is None:
            self._log.info(
                "Running command {}, current status {}".format(command, git_op.state)
            )
            git_op = self._git_op_execution_api._get_execution(
                git_op.repository.id, git_op.id
            )
            time.sleep(5)

        if git_op.success is False:
            raise GitException(
                "Git command failed with the message: {}".format(
                    git_op.command_result_message
                )
            )
        else:
            self._log.info("Git command {} finished".format(command))

        return git_op
